//	Zinc Application Framework - D_MOUSE.CPP
//	COPYRIGHT (C) 1990-1998.  All Rights Reserved.
//	Zinc Software Incorporated.  Pleasant Grove, Utah  USA

#include <z_dsp.hpp>
#include <z_evtmgr.hpp>
#include <z_i18n.hpp>
#include <z_image.hpp>
#include <z_mouse2.hpp>
#include <z_string.hpp> // For strcmp().
#include <z_sysevt.hpp>
#include <z_utils.hpp>

#define ZAF_MOUSE_INFO
#include "img_def.cpp"
#include "gbl_def.cpp"
#include "d_hware.hpp"

// ----- ZafMouse ----------------------------------------------------------

#if defined(ZAF_GRAPHICS_MODE)
ZafMouseData **ZafMouse::imageTable = ZAF_NULLP(ZafMouseData *);
#endif

void (* mouseQueueEvent)(unsigned, int *, int *) = ZAF_NULLF(void (*)(unsigned, int *, int *));
bool _devicePainting = false;
static bool _upToDate = true;
static int _horizontalRatio = 1;
static int _verticalRatio = 1;

#define MAX_EVENTS	20
static unsigned sFifo[MAX_EVENTS];
static int xFifo[MAX_EVENTS];
static int yFifo[MAX_EVENTS];
static int zState;
static int volatile in, out;

#if defined(DOS32PP) || (defined(_MSC_VER) && defined(DOS386))
extern void MouseRead();
#elif defined(DOSX286)
	bool _isOS2 = false;
#endif

#if defined(ZAF_MSDOS)
static ZafMouse *_mouse;
#endif

#if defined(ZAF_GRAPHICS_MODE)
// Define mouse system resources.
static struct SYS_MOUSE
{
	ZafDeviceImage imageID;
	ZafMouseStruct *image;
} ZAF_FARDATA _mouseImage[] =
{
	{ DM_VIEW, &_view_ZafMouse },
	{ DM_SELECT, &_view_ZafMouse },
	{ DM_EDIT, &_edit_ZafMouse },
	{ DM_WAIT, &_wait_ZafMouse },
	{ DM_MOVE, &_move_ZafMouse },
	{ DM_LEFT_SIDE, &_horizontal_ZafMouse },
	{ DM_RIGHT_SIDE, &_horizontal_ZafMouse },
	{ DM_TOP_SIDE, &_vertical_ZafMouse },
	{ DM_BOTTOM_SIDE, &_vertical_ZafMouse },
	{ DM_TOP_LEFT_CORNER, &_left_ZafMouse },
	{ DM_BOTTOM_RIGHT_CORNER, &_left_ZafMouse },
	{ DM_TOP_RIGHT_CORNER, &_right_ZafMouse },
	{ DM_BOTTOM_LEFT_CORNER, &_right_ZafMouse },
	{ DM_CROSS_HAIRS, &_crossHairs_ZafMouse },
	{ DM_CANCEL, &_cancel_ZafMouse },
	{ DM_DRAG_MOVE, &_dragMove_ZafMouse },
	{ DM_DRAG_COPY, &_dragCopy_ZafMouse },
	{ DM_DRAG_MOVE_MULTIPLE, &_dragMoveMultiple_ZafMouse },
	{ DM_DRAG_COPY_MULTIPLE, &_dragCopyMultiple_ZafMouse },
	{ DM_VERTICAL_SPLIT, &_vertical_split_ZafMouse },
	{ DM_HORIZONTAL_SPLIT, &_horizontal_split_ZafMouse },
	{ 0, 0 }
};
#endif

static void mouseQueueEventFunc(unsigned state, int *horizontalMickeys, int *verticalMickeys)
{
	static int i, deltaX, deltaY;

#if defined(DOS32PP) || (defined(_MSC_VER) && defined(DOS386))
	// Accumulate mickey count if less than ratios.
	static int xAmount, yAmount;
	if (state == zState &&
		ZafAbs(*horizontalMickeys + xAmount) < _horizontalRatio &&
		ZafAbs(*verticalMickeys + yAmount) < _verticalRatio)
	{
		xAmount += *horizontalMickeys;
		yAmount += *verticalMickeys;
		return;
	}
	else
	{
		*horizontalMickeys += xAmount;
		*verticalMickeys += yAmount;
		xAmount = *horizontalMickeys % _horizontalRatio;
		yAmount = *verticalMickeys % _verticalRatio;
	}
#endif
	deltaX = *horizontalMickeys / _horizontalRatio;
	deltaY = *verticalMickeys / _verticalRatio;

	i = in + 1;
	if (i >= MAX_EVENTS)
		i = 0;
	if (i != out && (state != zState || deltaX || deltaY))
	{
		// Update the mickey count.
		*horizontalMickeys %= _horizontalRatio;
		*verticalMickeys %= _verticalRatio;

		if (state == zState && in != out)
		{
			// Do an interrupt move of the mouse.
#if defined (DOSX286) && !defined(_MSC_VER)
			// Check for running under OS2.
			if (!_isOS2 && !state  && (deltaX || deltaY))
			{
				_mouse->MoveMouse(deltaX, deltaY);
				return;
			}
#elif defined (ZAF_MSDOS) && !defined(_MSC_VER) && !defined(__WATCOMC__) && !defined(DOS32PP)
			if (!state  && (deltaX || deltaY))
			{
# 	if defined(DOS16M)
#		if defined(ZAF_GRAPHICS_MODE)
				_mouse->MoveMouse(deltaX, deltaY);
#		endif
#	else
				_mouse->MoveMouse(deltaX, deltaY);
#	endif
				return;
			}
#endif
			// Update existing event in the queue, don't add a new one.
			xFifo[in] += deltaX;
			yFifo[in] += deltaY;
			return;
		}
		in = i;
		sFifo[i] = zState = state;
		xFifo[i] = deltaX;
		yFifo[i] = deltaY;
	}
}

void TranslateMouseEvent(ZafEventStruct &event)
{
	if (event.rawCode & S_SHIFT)
		event.modifiers |= S_SHIFT;
	if (event.rawCode & S_CTRL)
		event.modifiers |= S_CTRL;
	if (event.rawCode & S_ALT)
		event.modifiers |= S_ALT;
	event.rawCode &= ~(S_ALT | S_CTRL | S_SHIFT);
}

ZafMouse::ZafMouse(ZafDeviceState _state, ZafDeviceImage _imageType) :
	ZafDevice(E_MOUSE, _state), imageType(_imageType)
{
	if (!I_MouseOpen())	// Try to initialize the mouse driver.
		return;

#if defined(ZAF_MSDOS)
	_mouse = this;
#endif

	// Check the image information.
	ImageAllocate();
	SetStringID(className);
	SetNumberID((ZafNumberID)classID);

	// Initialize the mouse information.
	mouseQueueEvent = mouseQueueEventFunc;
	position.column = position.line = 0;
	offset.column = offset.line = 0;
#if defined(ZAF_GRAPHICS_MODE)
	position.coordinateType = offset.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
	position.coordinateType = offset.coordinateType = ZAF_CELL;
#endif
	installed = true;
}

ZafMouse::~ZafMouse(void)
{
	// Restore the old mouse information.
	if (installed)
	{
		mouseQueueEvent = ZAF_NULLF(void (*)(unsigned, int *, int *));
		I_MouseClose();
	}

	// Check the image information.
	ImageFree();

	installed = false;
}

ZafEventType ZafMouse::Event(const ZafEventStruct &event)
{
	// Make sure the mouse is installed.
	if (!installed)
		return (D_OFF);

	ZafDeviceState oldState = deviceState;
	ZafDeviceImage oldImage = imageType;
	ZafPositionStruct oldPosition = position;
	ZafEventType ccode = event.type;

	// Switch on the raw code.
	switch (ccode)
	{
	case D_INITIALIZE:
#if defined(ZAF_GRAPHICS_MODE)
		_horizontalRatio = 1;
		_verticalRatio = 1;
#elif defined(ZAF_TEXT_MODE)
		_horizontalRatio = display->columns == 40 ? 16 : 8;
		_verticalRatio = 8;
#endif
		position.column = display->columns / 2;
		position.line = display->lines / 2;
		oldImage = 0;
		oldState = D_OFF;

		// Initialize the original image.
		Event(imageType);
		break;

	case S_DEINITIALIZE:
		{
			ZafRegionStruct clip;

			// Start with screen region and pare it down.
			clip.left = clip.top = 0;
			clip.right = display->columns - 1;
			clip.bottom = display->lines - 1;
#if defined(ZAF_GRAPHICS_MODE)
			clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
			clip.coordinateType = ZAF_CELL;
#endif

			((ZafScreenDisplay *)display)->DeviceMove(ZAF_IMT_MOUSE_IMAGE, display->columns, display->lines, clip);
			deviceState = D_HIDE;
		}
		break;

	case D_POSITION:
		position = event.position;
		break;

	case DM_VIEW:
	case DM_SELECT:
	case DM_EDIT:
	case DM_WAIT:
	case DM_MOVE:
	case DM_LEFT_SIDE:
	case DM_RIGHT_SIDE:
	case DM_TOP_SIDE:
	case DM_BOTTOM_SIDE:
	case DM_TOP_LEFT_CORNER:
	case DM_TOP_RIGHT_CORNER:
	case DM_BOTTOM_LEFT_CORNER:
	case DM_BOTTOM_RIGHT_CORNER:
	case DM_CROSS_HAIRS:
	case DM_CANCEL:
	case DM_DRAG_MOVE:
	case DM_DRAG_COPY:
	case DM_DRAG_MOVE_MULTIPLE:
	case DM_DRAG_COPY_MULTIPLE:
	case DM_VERTICAL_SPLIT:
	case DM_HORIZONTAL_SPLIT:
		if (deviceState != D_OFF)
			deviceState = D_ON;
		imageType = ccode;
		break;

	case D_ON:
		{
			ZafRegionStruct clip;

			// Start with screen region and pare it down.
			clip.left = clip.top = 0;
			clip.right = display->columns - 1;
			clip.bottom = display->lines - 1;
#if defined(ZAF_GRAPHICS_MODE)
			clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
			clip.coordinateType = ZAF_CELL;
#endif

			((ZafScreenDisplay *)display)->DeviceMove(ZAF_IMT_MOUSE_IMAGE, position.column - offset.column,
				position.line - offset.line, clip);

			if (deviceState == D_OFF)
			{
				out = in;
				deviceState = D_ON;
			}
		}
		break;

	case D_OFF:
		deviceState = D_OFF;
		break;

	case D_STATE:
		return ((deviceState == D_OFF) ? D_OFF : imageType);

	case D_HIDE:
		{
			ZafRegionStruct clip;

			// Start with screen region and pare it down.
			clip.left = clip.top = 0;
			clip.right = display->columns - 1;
			clip.bottom = display->lines - 1;
#if defined(ZAF_GRAPHICS_MODE)
			clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
			clip.coordinateType = ZAF_CELL;
#endif

			((ZafScreenDisplay *)display)->DeviceMove(ZAF_IMT_MOUSE_IMAGE, display->columns, display->lines, clip);
			deviceState = event.type;
		}
		break;

	default:
		return (ccode);
	}

	// Reset the mouse information.
	if (oldImage == imageType && oldState == deviceState && oldPosition == position)
		return (ccode);
	if (oldPosition != position && deviceState != D_HIDE)
	{
		ZafRegionStruct clip;

		// Start with screen region and pare it down.
		clip.left = clip.top = 0;
		clip.right = display->columns - 1;
		clip.bottom = display->lines - 1;
#if defined(ZAF_GRAPHICS_MODE)
		clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
		clip.coordinateType = ZAF_CELL;
#endif

		_upToDate = ((ZafScreenDisplay *)display)->DeviceMove(ZAF_IMT_MOUSE_IMAGE,
			position.column - offset.column, position.line - offset.line, clip);
	}
	if (oldImage != imageType || oldState != deviceState)
	{
#if defined(ZAF_GRAPHICS_MODE)
		int tableOffset = 0;
		for (tableOffset = 0; imageTable[tableOffset]; tableOffset++)
			if (imageTable[tableOffset]->NumberID() == imageType)
				break;

		offset.column = imageTable[tableOffset]->hotSpotX;
		offset.line = imageTable[tableOffset]->hotSpotY;

		ZafRegionStruct clip;
		clip.left = clip.top = 0;
		clip.right = display->columns - 1;
		clip.bottom = display->lines - 1;
		clip.coordinateType = ZAF_PIXEL;

		_upToDate = ((ZafScreenDisplay *)display)->DeviceSet(ZAF_IMT_MOUSE_IMAGE,
			position.column - offset.column,
			position.line - offset.line, imageTable[tableOffset]->Width(),
			imageTable[tableOffset]->Height(), imageTable[tableOffset]->array, clip);
#elif defined(ZAF_TEXT_MODE)
		ZafRegionStruct clip;
		clip.left = clip.top = 0;
		clip.right = display->columns - 1;
		clip.bottom = display->lines - 1;
		clip.coordinateType = ZAF_CELL;

		_upToDate = ((ZafScreenDisplay *)display)->DeviceSet(ZAF_IMT_MOUSE_IMAGE, position.column,
			position.line, 1, 1, NULL, clip);
#endif
	}

	// Return the control code.
	return (ccode);
}

void ZafMouse::ImageAllocate(void)
{
#if defined(ZAF_GRAPHICS_MODE)
	// Check for an initialized mouse system.
	static bool initialized = false;
	if (initialized)
		return;

	// Initialize the mouse images.
	ZafI18nData::AddStaticModule(ImageFree);
	imageTable = new ZafMouseData *[sizeof(_mouseImage) / sizeof(SYS_MOUSE)];
	int i;
	for (i = 0; _mouseImage[i].imageID; i++)
	{
		imageTable[i] = new ZafMouseData(*_mouseImage[i].image);
		imageTable[i]->hotSpotX = _mouseImage[i].image->hotSpotX;
		imageTable[i]->hotSpotY = _mouseImage[i].image->hotSpotY;
		imageTable[i]->SetNumberID((ZafNumberID)_mouseImage[i].imageID);
	}
	imageTable[i] = ZAF_NULLP(ZafMouseData);

	initialized = true;
#endif
}

void ZafMouse::ImageFree(bool globalRequest)
{
#if defined(ZAF_GRAPHICS_MODE)
	// Destroy the static images.
	if (globalRequest && imageTable)
	{
		for (int i = 0; imageTable[i]; i++)
		{
			imageTable[i]->staticHandle = false;
			delete imageTable[i];
		}
		delete []imageTable;
		imageTable = ZAF_NULLP(ZafMouseData *);
	}
#endif
}

bool ZafMouse::IsA(ZafClassID compareID) const
{
	return ((compareID == classID) ? true : ZafDevice::IsA(compareID));
}

bool ZafMouse::IsA(ZafClassName compareName) const
{
	return (!strcmp(compareName, className) ? true : ZafDevice::IsA(compareName));
}

#if defined(ZAF_MSDOS)
void ZafMouse::MoveMouse(int deltaX, int deltaY)
{
	if (!_devicePainting)
	{
		_devicePainting = true;
		position.column += deltaX;
		position.line += deltaY;

		if (position.column >= display->columns)
			position.column = display->columns - 1;
		else if (position.column < 0)
			position.column = 0;
		if (position.line >= display->lines)
			position.line = display->lines - 1;
		else if (position.line < 0)
			position.line = 0;

		ZafRegionStruct clip;

		// Start with screen region and pare it down.
		clip.left = clip.top = 0;
		clip.right = display->columns - 1;
		clip.bottom = display->lines - 1;
#	if defined(ZAF_GRAPHICS_MODE)
		clip.coordinateType = ZAF_PIXEL;
#	elif defined(ZAF_TEXT_MODE)
		clip.coordinateType = ZAF_CELL;
#	endif

		((ZafScreenDisplay *)display)->DeviceMove(ZAF_IMT_CURSOR_IMAGE,
				display->columns, display->lines, clip);
		((ZafScreenDisplay *)display)->DeviceMove(ZAF_IMT_MOUSE_IMAGE,
				position.column - offset.column, position.line - offset.line, clip);
		_devicePainting = false;
	}
}
#endif

void ZafMouse::Poll(void)
{
	if (!installed)
		return;
#if defined(DOS32PP) || (defined(_MSC_VER) && defined(DOS386))
	MouseRead();
#endif
	// Make sure the mouse is valid.
	if (deviceState != D_OFF && in != out)
	{
		ZafEventStruct event;
		event.type = E_MOUSE;
		while (in != out)	// Get all mouse events.
		{
			// The converted member of event must be cleared because the
			// event structure is used multiple times in this loop, and
			// each time the event is processed converted is set.
			event.converted = ZAF_NULLP(ZafWindowObject);

			if (++out >= MAX_EVENTS)
				out = 0;
			event.rawCode = sFifo[out];
			position.column += xFifo[out];
			position.line   += yFifo[out];
			if (position.column >= display->columns)
				position.column = display->columns - 1;
			else if (position.column < 0)
				position.column = 0;
			if (position.line >= display->lines)
				position.line = display->lines - 1;
			else if (position.line < 0)
				position.line = 0;
			event.position = position;
			event.osEvent.type = E_MOUSE;
			event.osEvent.position.line = position.line;
			event.osEvent.position.column = position.column;
			TranslateMouseEvent(event);

			long line = position.line - lastClickPosition.line;
			long column = position.column - lastClickPosition.column;
			if (line < 0)
				line = -line;
			if (column < 0)
				column = -column;

			clock_t clickTime = clock();
			if (event.rawCode == (M_LEFT | M_LEFT_CHANGE) &&
				clickTime - lastClickTime  * 1000 / CLOCKS_PER_SEC <
				DoubleClickRate() && line < ZAF_DOS_DOUBLE_CLICK_DISTANCE &&
				column < ZAF_DOS_DOUBLE_CLICK_DISTANCE)
				event.rawCode |= S_DOUBLE_CLICK;
			else if (event.rawCode == (M_LEFT | M_LEFT_CHANGE) )
			{
				lastClickTime = clickTime;
				lastClickPosition = position;
			}

			eventManager->Put(event);
		}
		if (deviceState != D_HIDE)
		{
			ZafRegionStruct clip;

			// Start with screen region and pare it down.
			clip.left = clip.top = 0;
			clip.right = display->columns - 1;
			clip.bottom = display->lines - 1;
#if defined(ZAF_GRAPHICS_MODE)
			clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
			clip.coordinateType = ZAF_CELL;
#endif

			_upToDate = ((ZafScreenDisplay *)display)->DeviceMove(ZAF_IMT_MOUSE_IMAGE,
				position.column - offset.column,
				position.line - offset.line, clip);
		}
	}
	// Mouse could not be updated during a previous paint operation.
	else if (!_upToDate)
	{
		ZafDeviceImage newImage = imageType;
		imageType = 0;
		Event(ZafEventStruct(newImage));
		_upToDate = true;
	}
}

// ----- Attributes ---------------------------------------------------------

ZafDeviceImage ZafMouse::SetImageType(ZafDeviceImage tImageType)
{
	imageType = tImageType;
	return (imageType);
}

int ZafMouse::SetDoubleClickRate(int tDoubleClickRate)
{
	doubleClickRate = tDoubleClickRate;
	return (doubleClickRate);
}
