//	Zinc Application Framework - D_VLIST.CPP
//	COPYRIGHT (C) 1990-1998.  All Rights Reserved.
//	Zinc Software Incorporated.  Pleasant Grove, Utah  USA

#include <time.h>
#include <z_combo.hpp>
#define USE_RAW_KEYS
#include <z_keymap.hpp>
#include <z_mouse2.hpp>
#include <z_scrll2.hpp>
#include <z_str1.hpp>
#include <z_string.hpp>
#include <z_utils.hpp>
#include <z_vlist.hpp>

// ----- ZafVtList --------------------------------------------------------

ZafWindowObject *ZafVtList::Add(ZafWindowObject *object, ZafWindowObject *position)
{
	// Mark the object as view only.
	ZafString *string = DynamicPtrCast(object, ZafString);
	if (string)
		string->SetViewOnly(true);
	if (OSDraw())
		object->SetBordered(false);
	if (!object->SupportObject())
	{
		object->SetSystemObject(false);
		object->SetParentPalette(true);
		object->SetParentDrawBorder(true);
		object->SetParentDrawFocus(true);
	}
	ZafWindow::Add(object, position);
	if (screenID)
		Event(N_SIZE);

	// Return the added object.
	return (object);
}

ZafRegionStruct ZafVtList::ConvertToDrawRegion(const ZafWindowObject *object,
	const ZafRegionStruct *region) const
{
	return (ZafWindowObject::ConvertToDrawRegion(object, region));
}

ZafEventType ZafVtList::DispatchMouseEvent(const ZafEventStruct &event)
{
	return (ZafWindowObject::DispatchMouseEvent(event));
}

ZafWindowObject *ZafVtList::FindMouseObject(const ZafEventStruct &event)
{
	return (ZafWindowObject::FindMouseObject(event));
}

ZafEventType ZafVtList::DragDropEvent(const ZafEventStruct &event)
{
	ZafEventType ccode = event.type;

	switch (ccode)
	{
	case S_DROP_DEFAULT:
		ccode = S_DROP_MOVE;
		// Fall through to S_DROP_MOVE.
	case S_DROP_COPY:
	case S_DROP_MOVE:
		if (Disabled() || !AcceptDrop() || windowManager->dragObject == this)
			return (S_ERROR);
		else
		{
			// Clear all selections if extended selection list.
			if (selectionType == ZAF_EXTENDED_SELECTION)
			{
				for (ZafWindowObject *item = First(); item; item = item->Next())
					item->SetSelected(false);
			}

			if (windowManager->dragObject->IsA(ID_ZAF_LIST))
			{
				ZafWindow *dragList = DynamicPtrCast(windowManager->dragObject, ZafWindow);

				bool saveAutoUpdateSource = dragList->AutomaticUpdate();
				bool saveAutoUpdateTarget = automaticUpdate;

				if (ccode == S_DROP_MOVE)
					dragList->SetAutomaticUpdate(false);
				SetAutomaticUpdate(false);

				// Move the selected objects.
				ZafWindowObject *object;
				if (dragList->IsA(ID_ZAF_COMBO_BOX))
					object = DynamicPtrCast(dragList, ZafComboBox)->First();
				else
					object = dragList->First();
				while (object)
				{
					ZafWindowObject *temp = object->Next();
					if (object->Selected())
					{
						if (ccode == S_DROP_COPY)
							Add(object->Duplicate());
						else
						{
							dragList->Subtract(object);
							Add(object);
						}
					}
					object = temp;
				}
				if (ccode == S_DROP_MOVE)
					dragList->SetAutomaticUpdate(saveAutoUpdateSource);
				SetAutomaticUpdate(saveAutoUpdateTarget);
			}
			else
			{
				const ZafIChar *dropText = windowManager->dragObject->Text();
				if (dropText && *dropText)
				{
					ZafString *newChild = new ZafString(0, 0, 0, new ZafStringData(dropText));
					newChild->SetBordered(false);
					Add(newChild);
					if (ccode != S_DROP_COPY)
						windowManager->dragObject->SetText(ZafLanguageManager::blankString);
				}
				else
					ccode = S_ERROR;
			}
		}
		break;

	default:
		ccode = ZafWindow::DragDropEvent(event);
		break;
	}

	return (ccode);
}

#if defined(ZAF_DEFAULT_DRAW)
ZafEventType ZafVtList::Draw(const ZafEventStruct &event, ZafEventType ccode)
{
	// Compute the actual draw region.
	ZafPaletteState state = PaletteState();
	ZafRegionStruct drawRegion = BeginDraw();

	display->SetClipRegion(event.region);

	// Redraw the list background if necessary.
	if (!last || Last()->zafRegion.bottom < clientRegion.bottom)
	{
		int top = drawRegion.top;
		if (last)
			top = Last()->zafRegion.bottom;
		display->SetPalette(LogicalPalette(ZAF_PM_LIGHT_SHADOW, state));
		display->Rectangle(drawRegion.left, top,
			drawRegion.right, drawRegion.bottom, 0, true);
	}

	// Don't let the children do nested BeginDraw() calls.
	EndDraw();

	// Draw the child objects, since they are !systemObject.
	for (ZafWindowObject *object = First(); object; object = object->Next())
	{
		ZafEventStruct childEvent = event;
		if (object->zafRegion.Overlap(event.region, childEvent.region))
		{
			childEvent.region.HzShift(-object->zafRegion.left);
			childEvent.region.VtShift(-object->zafRegion.top);
			object->Draw(childEvent, ccode);
		}
	}

	return (ccode);
}

ZafEventType ZafVtList::DrawBorder(ZafRegionStruct &region, ZafEventType ccode)
{
	if (temporary && parent && parent->IsA(ID_ZAF_COMBO_BOX))
	{
		ZafPaletteStruct lightShadow = LogicalPalette(ZAF_PM_LIGHT_SHADOW, ZAF_PM_ANY_STATE);
		ZafPaletteStruct darkShadow = LogicalPalette(ZAF_PM_DARK_SHADOW, ZAF_PM_ANY_STATE);
		ZafPaletteStruct whiteShadow = lightShadow;
		ZafPaletteStruct blackShadow = darkShadow;

		// Copy the background colors to the foreground since display->Rectangle()
		// uses the foreground color.  The "shadow" palettes have different colors
		// in the foreground and background entries so they can generate two shades
		// of shadow.
		whiteShadow.colorForeground = whiteShadow.colorBackground;
		blackShadow.colorForeground = blackShadow.colorBackground;

		display->SetPalette(blackShadow);
		display->Rectangle(region);
		region--;

		display->SetPalette(whiteShadow);
		display->Rectangle(region);
		region--;

		return (ccode);
	}
	else
		return (ZafWindow::DrawBorder(region, ccode));
}

ZafEventType ZafVtList::DrawFocus(ZafRegionStruct &region, ZafEventType ccode)
{
	display->SetPalette(*((ZafScreenDisplay *)display)->xorPalette);
	display->SetMode(ZAF_MODE_XOR);
	display->Rectangle(region, -1);
	display->SetMode(ZAF_MODE_COPY);
	return (ccode);
}
#endif

ZafEventType ZafVtList::Event(const ZafEventStruct &event)
{
	static clock_t firstDown = 0, secondDown = 0, currentTime = clock();
	static ZafWindowObject *lastSelected = ZAF_NULLP(ZafWindowObject);

	ZafEventType ccode = LogicalEvent(event);

	// Check for zinc events.
	ZafWindowObject *object;
	switch (ccode)
	{
	// ----- Create messages -----
	case S_INITIALIZE:
		ccode = ZafWindow::Event(event);
		firstVisible = lastVisible = ZAF_NULLP(ZafWindowObject);
#if defined(ZAF_TEXT_MODE)
		if (temporary)
			SetBordered(true);
#endif
		break;

	case S_HOT_KEY:
		ccode = S_UNKNOWN;
		break;

	case S_REGISTER_OBJECT:
		// Register the object. By registering children here, we prevent them
		// from being labelled as system objects. This will cause them to
		// get their palettes from the list and to let the list
		// draw their focus.
		RegisterObject();

		for (object = SupportFirst(); object; object = object->Next())
			object->Event(event);

		// Define the children to have the same screenID as their parent.
		for (object = First(); object; object = object->Next())
			object->screenID = screenID;
		break;

	// ----- Position & Size messages -----
	case S_COMPUTE_SIZE:
		ccode = ZafWindow::Event(event);

		// Define a screen region if temporary (being used for a combo box).
		if (temporary)
			((ZafScreenDisplay *)display)->RegionDefine(screenID, zafRegion);
		break;

	case N_SIZE:
		{
			// width and height are the default sizes of each item in the list.
			// If the list is !osDraw, the size of each object is determined
			// at run-time.
			int width = clientRegion.right - clientRegion.left;
			int height = display->cellHeight - display->preSpace - display->postSpace;

			// top and bottom are the bounding areas of the list's client region.
			// line is the current position of the specified object in the list.
			int top = 0;
			int bottom = clientRegion.bottom - clientRegion.top;

			// Position all objects in the list. This algorithm traverses the
			// list by first moving from the first visible and positioning each
			// object, then by moving to all object's above the firstVisible
			// to position them within the list. The first execution sets
			// firstVisible as the first item in the list.
			if (!firstVisible)
				firstVisible = First();

			// Work down from firstVisible.
			int line = top;
			object = lastVisible = firstVisible;
			for ( ; object; object = object->Next())
			{
				// Compute the desired size.
				object->Event(S_COMPUTE_SIZE);

				// Reset the object's zafRegion.
				if (!OSDraw())
					height = object->zafRegion.bottom - object->zafRegion.top;
				object->zafRegion.left = 0;
				object->zafRegion.right = width;
				object->zafRegion.top = line;

#if defined(ZAF_TEXT_MODE)
				// Added -1 to correct size of list items in text mode.
				object->zafRegion.bottom = line + height - 1;
#else
				object->zafRegion.bottom = line + height;
#endif
				line = object->zafRegion.bottom + 1;

				// Check for lastVisible and manage the scroll values.
				if (object->zafRegion.bottom <= bottom)
					lastVisible = object;
			}

			// Work up from firstVisible.
			if (firstVisible)
			{
				line = top - 1;
				for (object = firstVisible->Previous(); object; object = object->Previous())
				{
					// Compute the desired size.
					object->Event(S_COMPUTE_SIZE);

					// Reset the object's zafRegion.
					if (!OSDraw())
						height = object->zafRegion.bottom - object->zafRegion.top;
					object->zafRegion.left = 0;
					object->zafRegion.right = width;
					object->zafRegion.bottom = line;
					object->zafRegion.top = line - height;
					line = object->zafRegion.top + - 1;
				}
			}

			// Finalize the sizing operation.
			for (object = First(); object; object = object->Next())
				object->OSSize();

			// Adjust the scroll information.
			ScrollEvent(S_VSCROLL_COMPUTE);
		}
		break;

	// ----- Selection messages -----
	case L_SELECT:
		{
		ZafWindowObject *currentObject = Current();
		if (!currentObject)
			break;

		if (event.key.value != ' ')
		{
			ccode = S_UNKNOWN;
			break;
		}

		if (SelectionType() != ZAF_SINGLE_SELECTION)
			currentObject->ToggleSelected();
		else
			currentObject->SetSelected(true);

		// Call the associated user function.
		(currentObject->*(currentObject->memberUserFunction))(event, ccode);

		if (temporary)
			windowManager->Event(S_CLOSE_TEMPORARY);
		}
		break;

	case L_BEGIN_SELECT:
	case L_END_SELECT:
		if (editMode)
		{
			ZafWindowObject::Event(event);
			break;
		}
		currentTime = clock();

		// Continue to L_CONTINUE_SELECT.
	case L_CONTINUE_SELECT:
//	case L_CONTINUE_COPY_DRAG:
//	case L_END_COPY_DRAG:
	case L_VIEW:
		{
			if (Draggable() && ccode != L_VIEW && !windowManager->dragObject)
				ZafWindowObject::Event(event);
			if (editMode || windowManager->dragObject)
			{
				ZafWindowObject::Event(event);
				break;
			}

			// Process L_VIEW only if combo-box drop-down list.
			if (ccode == L_VIEW && !temporary)
				return (ZafWindowObject::Event(event));

			// Check for the current focus.
			if (ccode != L_BEGIN_SELECT && !Focus() && !temporary)
				break; // illegal selection process.

			// Check for out-of-bounds.
			ZafPositionStruct mousePosition = event.position;
			ZafRegionStruct windowClient = clientRegion;
			// event.position is 0,0 based for the visible client region after 
			// it has been dispatched.
			windowClient.HzShift(-windowClient.left);
			windowClient.VtShift(-windowClient.top);
			if (!windowClient.Overlap(event.position))
			{
				// Check if over the scroll bar, since this is an atom.
				if (SupportFirst() && SupportFirst()->zafRegion.Overlap(event.position))
					return (SupportFirst()->DispatchMouseEvent(event));
				else
					return (ZafWindowObject::Event(event));
			}

/* START BLOCK COMMENT
**				if (ccode == L_BEGIN_SELECT && Draggable())
**					windowManager->dragObject = this;
END BLOCK COMMENT */

/* START BLOCK COMMENT
**				// Adjust for client region.
**				mousePosition.column -= clientRegion.left;
**				mousePosition.line -= clientRegion.top;
END BLOCK COMMENT */

			// Try to find an overlapping item.
			for (object = firstVisible; object; object = object->Next())
				if (object->zafRegion.Overlap(mousePosition))
				{
					if (object->Disabled())
						return (ccode);

					if (current != object)
					{
						object->SetFocus(true);

						// Change the selected object.
						if (SelectionType() == ZAF_SINGLE_SELECTION)
							object->SetSelected(true);
					}
					break;
				}

			// If the mouse did not overlap any objects, break.
			if (!object)
				break;

			if (ccode == L_BEGIN_SELECT)
			{
				if (lastSelected != object)
				{
					// place current time in secondDown.  If put it in firstDown
					// a double-click may be interpreted even though this was the
					// first click on the new object.  If a double-click is being
					// initiated secondDown gets shifted to firstDown on the
					// next up click.
					firstDown = 0;
					secondDown = currentTime;
					lastSelected = object;
				}
				else
					secondDown = currentTime;
			}

			// Check for a new selection.
			if (ccode == L_END_SELECT)
			{
				ZafEventStruct uEvent = event;
				uEvent.type = L_SELECT;

				// Change the selection.  Only toggle if not a double-click.
				// Toggling on a double-click would switch the object back to
				// the state it was in before the double-click started.
				if (SelectionType() != ZAF_SINGLE_SELECTION)
				{
					if (uEvent.type != L_DOUBLE_CLICK)
						object->ToggleSelected();
				}
				else
					object->SetSelected(true);

				// Call the associated user function.
				(object->*(object->memberUserFunction))(uEvent, uEvent.type);

				// Notify the parent if a temporary window.
				if (Temporary())
					windowManager->Event(S_CLOSE_TEMPORARY);
			}
		}
		break;

	case S_CURRENT:
		if (current && SelectionType() == ZAF_SINGLE_SELECTION)
			Current()->SetSelected(true);
		ccode = ZafWindow::Event(event);
		break;

	case S_NON_CURRENT:
		if (Temporary() && current)
			Current()->Event(event);
		ccode = ZafWindow::Event(event);
		break;

	case E_KEY:
		if (temporary && parent)
			ccode = parent->Event(event);
		else
			ccode = ZafWindow::Event(event);
		break;

	// ----- Movement messages -----
	case L_NEXT:
	case L_PREVIOUS:
		ccode = S_UNKNOWN;
		break;

	case L_UP:
	case L_DOWN:
	case L_FIRST:
	case L_LAST:
	case L_PGDN:
	case L_PGUP:
		ccode = MoveEvent(ccode);
		break;

	// ----- Default operation -----
	default:
		ccode = ZafWindow::Event(event);
		break;
	}

	// Return the control code.
	return (ccode);
}

void ZafVtList::GetClip(ZafWindowObject *object, ZafRegionStruct &clip,
	ZafRegionStruct &virtualRegion)
{
	if (!Temporary())
		ZafWindow::GetClip(object, clip, virtualRegion);
	else
	{
		// Adjust the virtualRegion region.
		virtualRegion.left += zafRegion.left;
		virtualRegion.top += zafRegion.top;
		virtualRegion.right += zafRegion.left;
		virtualRegion.bottom += zafRegion.top;

		if (virtualRegion.left > clip.left)
			clip.left = virtualRegion.left;
		if (virtualRegion.top > clip.top)
			clip.top = virtualRegion.top;
		clip.right = ZafMin(clip.right, clip.left + zafRegion.Width() - 1);
		clip.bottom = ZafMin(clip.bottom, clip.top + zafRegion.Height() - 1);

		// Adjust for the client region if the object is in the client area.
		if (object != this && !object->SupportObject())
		{
			virtualRegion.left += clientRegion.left;
			virtualRegion.top += clientRegion.top;
			virtualRegion.right += clientRegion.left;
			virtualRegion.bottom += clientRegion.top;

			if (virtualRegion.left > clip.left)
				clip.left = virtualRegion.left;
			if (virtualRegion.top > clip.top)
				clip.top = virtualRegion.top;
			clip.right = ZafMin(clip.right, clip.left + clientRegion.Width() - 1);
			clip.bottom = ZafMin(clip.bottom, clip.top + clientRegion.Height() - 1);
		}
	}
}

ZafEventType ZafVtList::MoveEvent(const ZafEventStruct &event)
{
	// Determine the new selection.
	ZafEventType ccode = event.type;
	ZafWindowObject *object = Current();
	ZafWindowObject *pObject = object ? object->Previous() : ZAF_NULLP(ZafWindowObject);
	ZafWindowObject *nObject = object ? object->Next() : ZAF_NULLP(ZafWindowObject);

	switch (ccode)
	{
	case L_UP:
		if (pObject)
			object = pObject;
		break;

	case L_DOWN:
		if (nObject)
			object = nObject;
		break;

	case L_PGUP:
		// Check for page actions.
		if (!object || !pObject)
			break; // No current object exists.
		else if (pObject->zafRegion.top < 0)
		{
			// Go to the top of the previous page.
			int height = clientRegion.Height() - object->zafRegion.Height();
			while (height > 0 && object->Previous())
			{
				object = object->Previous();
				height -= object->zafRegion.Height() - ZAF_BORDER_SIZE;
			}
		}
		else
		{
			// Go to the top of the current page.
			while (object->Previous() && object->Previous()->zafRegion.top >= 0)
				object = object->Previous();
		}
		break;

	case L_PGDN:
		// Check for page actions.
		if (!object || !nObject)
			break; // No current object exists.
		else if (nObject->zafRegion.bottom > clientRegion.Height())
		{
			// Go to the bottom of the next page.
			int height = clientRegion.Height() - object->zafRegion.Height();
			while (height > 0 && object->Next())
			{
				object = object->Next();
				height -= object->zafRegion.Height() - ZAF_BORDER_SIZE;
			}
		}
		else
		{
			// Go to the bottom of the current page.
			while (object->Next() && object->Next()->zafRegion.bottom <= clientRegion.Height())
				object = object->Next();
		}
		break;

	case L_FIRST:
		object = First();
		break;

	case L_LAST:
		object = Last();
		break;
	}

	// Check the current object.
	if (object != current)
	{
		// Reset the current object.
		object->SetFocus(true);

		// Check the scroll information.
		ScrollEvent(S_VSCROLL_CHECK);

		// Change the selected object.
		if (SelectionType() == ZAF_SINGLE_SELECTION)
		{
			object->SetSelected(true);

			// Call the associated user function.
			(object->*(object->memberUserFunction))(event, L_SELECT);
		}
	}
	return (ccode);
}

ZafWindowObject *ZafVtList::NotifyFocus(ZafWindowObject *focusObject, bool setFocus)
{
	return (ZafWindow::NotifyFocus(focusObject, setFocus));
}

ZafWindowObject *ZafVtList::NotifySelection(ZafWindowObject *newObject, bool selected)
{
	ZafWindow::NotifySelection(newObject, selected);

	// If a combo-box drop-down list and the new object is selected,
	// notify the combo-box.
	if (temporary && parent && parent->IsA(ID_ZAF_COMBO_BOX) && newObject->Selected())
	{
		SetCurrent(newObject);
		DynamicPtrCast(parent, ZafWindow)->NotifySelection(newObject, true);
	}
	return (newObject);
}

ZafEventType ZafVtList::ScrollEvent(const ZafEventStruct &event)
{
	ZafWindowObject *object;
	int top = 0;
	int bottom = clientRegion.bottom - clientRegion.top;

	if (!screenID)
		return (S_UNKNOWN);

	// Process the scrolling request.  Scrolling at the vertical list level
	// is based on the number of list items.
	ZafEventType ccode = event.type;
	switch (ccode)
	{
	case S_VSCROLL_CHECK:
		// Check for a valid object.
		object = Current();
		if (!object)
			return (S_ERROR);

		// Scroll the object into view.
		{
		ZafScrollStruct vScroll;
		vScroll.delta = 0;
		if (object->zafRegion.top < firstVisible->zafRegion.top)
		{
			// Determine the downward scrolling delta.
			ZafWindowObject *temp = object;
			while (temp && temp != firstVisible)
			{
				vScroll.delta--;
				temp = temp->Next();
			}
		}
		else if (object->zafRegion.bottom > lastVisible->zafRegion.bottom)
		{
			int difference = object->zafRegion.bottom -
				(clientRegion.bottom - clientRegion.top);

			// Determine the upward scrolling delta.
			ZafWindowObject *temp = firstVisible;

			// firstVisible is at the top of the list, so we need to find an 
			// object whose zafRegion.top is greater than the difference needed
			// to scroll the full current object into view.
			while (temp && difference - temp->zafRegion.top >= 0)
			{
				vScroll.delta++;
				temp = temp->Next();
			}
		}

		// Scroll the list elements or adjust the values.
		if (vScroll.delta)
			ScrollEvent(ZafEventStruct(S_VSCROLL, 0, vScroll));
		}
		break;

	case S_VSCROLL:
	case N_VSCROLL:
		// Determine the actual scrolling boundary, based on the size of 
		// each list element.
		{
		ZafWindowObject *newFirst = firstVisible;
		ZafWindowObject *newLast = lastVisible;
		if (event.scroll.delta)
		{
			if (event.scroll.delta > 0)
			{
				if (lastVisible == Last())
					break;

				// Scroll downward by adjusting newFirst as long as there is 
				// another newLast to scroll to.
				for (int i = (int)event.scroll.delta; i > 0; i--)
					if (newLast)
					{
						newFirst = newFirst->Next();
						newLast = newLast->Next();
					}
					else
						break;
			}
			else if (event.scroll.delta < 0)
			{
				if (firstVisible == First())
					break;

				// Scroll upward by first adjusting newFirst.
				for (int i = (int)event.scroll.delta; i < 0; i++)
					if (newFirst->Previous())
						newFirst = newFirst->Previous();
					else
						break;
			}

			if (!newFirst)
				newFirst = Last();

			// Now finish by computing newLast.
			newLast = newFirst;
			while (newLast->Next())
				if (newLast->Next()->zafRegion.bottom - newFirst->zafRegion.top <= bottom - top)
					newLast = newLast->Next();
				else
					break;
		}
		else
			return (S_ERROR);

		// Redisplay the list to get final positioning.
		if (newFirst != firstVisible || newLast != lastVisible)
		{
			// Adjust the first and last visible.
			firstVisible = newFirst;
			lastVisible = newLast;

			// Prevent objects from drawing when they get the OSSize().
			SetAutomaticUpdate(false);

			// Adjust all of the child positions.
			int yDelta = firstVisible->zafRegion.top;
			for (object = First(); object; object = object->Next())
			{
				object->zafRegion.top -= yDelta;
				object->zafRegion.bottom -= yDelta;
				object->OSSize();
			}

			// Allow the list to update its client area all at once.
			SetAutomaticUpdate(true);

			// Adjust the scroll values, if appropriate.
			if (ccode == S_VSCROLL)
			{
				ZafScrollBar *scrollBar = VerticalScrollBar();
				if (scrollBar)
					scrollBar->Event(ZafEventStruct(N_VSCROLL, 0, event.scroll));
			}
		}
		}
		break;

	case S_VSCROLL_COMPUTE:
		// Reset the scroll values.
		{
		ZafScrollStruct vScroll;
		vScroll.delta = 1;
		vScroll.minimum = vScroll.maximum = 0;
		vScroll.showing = vScroll.current = 0;

		// Work down from firstVisible.
		for (object = firstVisible; object; object = object->Next())
		{
			if (object->zafRegion.bottom <= bottom)
				vScroll.showing++;
			vScroll.maximum++;
		}

		// Work up from firstVisible.
		if (firstVisible)
		{
			for (object = firstVisible->Previous(); object; object = object->Previous())
			{
				vScroll.current++;
				vScroll.maximum++;
			}
		}

		vScroll.maximum -= vScroll.showing;

		// Adjust the vertical scroll-bar values.
		ZafScrollBar *scrollBar = VerticalScrollBar();
		if (scrollBar)
			scrollBar->Event(ZafEventStruct(S_VSCROLL_SET, 0, vScroll));
		}
		break;

	default:
		// No other messages apply to the vertical list object.
		return (S_ERROR);
	}

	// Return the control code.
	return (ccode);
}

ZafWindowObject *ZafVtList::Subtract(ZafWindowObject *object)
{
	// Remove the object from the list.
	if (object == firstVisible)
		firstVisible = object->Previous();
	ZafWindowObject *retValue = ZafWindow::Subtract(object);

	if (screenID)
		Event(N_SIZE);

	return (retValue);
}

bool ZafVtList::SetAutomaticUpdate(bool setAutomaticUpdate)
{
	// Defer to the base class.
	return (ZafWindow::SetAutomaticUpdate(setAutomaticUpdate));
}

bool ZafVtList::SetVisible(bool tVisible)
{
	return (ZafWindow::SetVisible(tVisible));
}

// ----- OS Specific Functions ----------------------------------------------

void ZafVtList::OSDestroy(void)
{
	// Notify each of the children.
	for (ZafWindowObject *object = First(); object; object = object->Next())
		object->Event(S_DEINITIALIZE);

	firstVisible = lastVisible = ZAF_NULLP(ZafWindowObject);
}

ZafEventType ZafVtList::OSEvent(const ZafEventStruct &)
{
	return (S_ERROR);
}

void ZafVtList::OSMapPalette(ZafPaletteStruct &palette, ZafPaletteType type, ZafPaletteState state)
{
	ZafWindowObject::OSMapPalette(palette, type, state);
}

void ZafVtList::OSRegisterObject(void)
{
	ZafWindow::OSRegisterObject();
}

OSWindowID ZafVtList::OSScreenID(ZafScreenIDType) const
{
	return (screenID);
}

void ZafVtList::OSSize(void)
{
	ZafWindow::OSSize();
//	clientRegion.left++;
}

void ZafVtList::OSSort(void)
{
}

ZafError ZafVtList::OSUpdatePalettes(ZafPaletteType zafTypes, ZafPaletteType osTypes)
{
	return (ZafWindow::OSUpdatePalettes(zafTypes, osTypes));
}

