//	Zinc Application Framework - D_WINOBJ.CPP
//	COPYRIGHT (C) 1990-1998.  All Rights Reserved.
//	Zinc Software Incorporated.  Pleasant Grove, Utah  USA

#include <time.h>
#include <z_ctype.hpp>
#include <z_error.hpp>
#include <z_help.hpp>
#define USE_RAW_KEYS
#include <z_keymap.hpp>
#include <z_mouse2.hpp>	// For DM_VIEW.
#include <z_stdlib.hpp>
#include <z_string.hpp>
#include <z_utils.hpp>
#include <z_win.hpp>

// ----- ZafWindowObject ---------------------------------------------------

ZafRegionStruct ZafWindowObject::BeginDraw(void)
{
	// Start with screen region and pare it down.
	ZafRegionStruct clip, virtualRegion;
	clip.left = clip.top = 0;
	clip.right = display->columns - 1;
	clip.bottom = display->lines - 1;
#if defined(ZAF_GRAPHICS_MODE)
	clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
	clip.coordinateType = ZAF_CELL;
#endif

	// Start with the object's region, based at (0, 0), and translate it
	// based on its parents' regions.
	virtualRegion.left = virtualRegion.top = 0;
	virtualRegion.right = zafRegion.Width() - 1;
	virtualRegion.bottom = zafRegion.Height() - 1;
	virtualRegion.coordinateType = zafRegion.coordinateType;

	// Get the clip and virtual regions.
	GetClip(this, clip, virtualRegion);

	// Initialize the display class for drawing.
	display->BeginDraw(0, screenID, virtualRegion, clip);

	// Shift the region to be (0, 0)-based.
	return (ConvertToDrawRegion(this, &virtualRegion));
}

ZafRegionStruct ZafWindowObject::ConvertToDrawRegion(const ZafWindowObject *object, const ZafRegionStruct *tZafRegion) const
{
	// Start with the native zaf region.
	ZafRegionStruct convert = tZafRegion ? *tZafRegion : object->zafRegion;
	if (object == this)
	{
		// Shift the region to be (0, 0)-based for drawing.
		convert.right -= convert.left;
		convert.bottom -= convert.top;
		convert.left = convert.top = 0;
	}

	// Return the result.
	return (convert);
}

ZafPositionStruct ZafWindowObject::ConvertToObjectPosition(const ZafPositionStruct screenPosition) const
{
	ZafPositionStruct zafPosition = screenPosition;
	if (parent && windowManager->Index(this) == -1)
		zafPosition = parent->ConvertToZafPosition(this, &zafPosition);
	else
		zafPosition = windowManager->ConvertToZafPosition(this, &zafPosition);

	zafPosition.line -= zafRegion.top;
	zafPosition.column -= zafRegion.left;

	return (zafPosition);
}

ZafPositionStruct ZafWindowObject::ConvertToOSPosition(const ZafWindowObject *,
	const ZafPositionStruct *zafPosition) const
{
	ZafPositionStruct osPosition;

	if (zafPosition)
	{
		osPosition.column = zafPosition->column + zafRegion.left;
		osPosition.line = zafPosition->line + zafRegion.top;
		osPosition.coordinateType = zafPosition->coordinateType;
	}
	else
	{
		osPosition.column = osPosition.line = 0;
#if defined(ZAF_GRAPHICS_MODE)
		osPosition.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
		osPosition.coordinateType = ZAF_CELL;
#endif
	}

	if (parent)
		osPosition = parent->ConvertToOSPosition(this, &osPosition);

	return (osPosition);
}

ZafPositionStruct ZafWindowObject::ConvertToScreenPosition(const ZafPositionStruct objectPosition) const
{
	ZafPositionStruct osPosition;
/* START BLOCK COMMENT
**		if (parent)
**			osPosition = parent->ConvertToOSPosition(this, &objectPosition);
**		else
**			osPosition = windowManager->ConvertToOSPosition(this, &objectPosition);
END BLOCK COMMENT */
		osPosition = ConvertToOSPosition(0, &objectPosition);

	return (osPosition);
}

bool ZafWindowObject::ConvertToZafEvent(ZafEventStruct &event)
{
	event.converted = this;
	if (event.osEvent.type == E_MOUSE)
	{
		event.position.line = event.osEvent.position.line;
		event.position.column = event.osEvent.position.column;
		event.position = ConvertToObjectPosition(event.position);
	}

	return (event.converted ? true : false);
}

ZafPositionStruct ZafWindowObject::ConvertToZafPosition(const ZafWindowObject *,
	const ZafPositionStruct *osPosition) const
{
	ZafPositionStruct zafPosition;

	if (osPosition)
	{
		zafPosition.column = osPosition->column - zafRegion.left;
		zafPosition.line = osPosition->line - zafRegion.top;
		zafPosition.coordinateType = osPosition->coordinateType;
	}
	else
	{
		zafPosition.column = zafPosition.line = 0;
#if defined(ZAF_GRAPHICS_MODE)
		zafPosition.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
		zafPosition.coordinateType = ZAF_CELL;
#endif
	}

	if (parent)
		zafPosition = parent->ConvertToZafPosition(this, &zafPosition);

	return (zafPosition);
}

ZafEventType ZafWindowObject::DispatchMouseEvent(const ZafEventStruct &event)
{
	if (modifyMode)
		return (Modify(event));

	if (windowManager->MouseEventRoute() == this)
		return (Event(event));

	// LogicalEvent() must be called first to map the logical events (L_*) and
	// to convert the event to the context of this object.
	ZafEventType ccode = LogicalEvent(event);

	// An object being clicked on receives focus before processing
	// the selection event.
	if (ccode == L_BEGIN_SELECT && !Disabled() && !Noncurrent())
		SetFocus(true);

	// Modify() is called from ZafWindowObject::Event().
	if (editMode && (ccode == L_BEGIN_SELECT || ccode == L_END_SELECT))
		return (ZafWindowObject::Event(event));

	if (ccode == L_VIEW && this != windowManager->mouseObject)
	{
		// Generate the N_MOUSE_ENTER and N_MOUSE_LEAVE messages.
		if (windowManager->mouseObject)
			windowManager->mouseObject->Event(N_MOUSE_LEAVE);
		if (Disabled())
			windowManager->mouseObject = ZAF_NULLP(ZafWindowObject);
		else
		{
			// Set the mouse object.
			windowManager->mouseObject = this;

			// Create a mouse enter event that contains event.position.
			ZafEventStruct enterEvent = event;
			enterEvent.type = N_MOUSE_ENTER;
			Event(enterEvent);
		}
	}

	// Dispatch all mouse events to the leaf-most object.
	if (Disabled())
		ccode = S_UNKNOWN;
	else
		ccode = Event(event);

	return (ccode);
}

ZafWindowObject *ZafWindowObject::FindMouseObject(const ZafEventStruct &)
{
	// If the object is disabled return a null pointer.
	if (Disabled())
		return (ZAF_NULLP(ZafWindowObject));
	return (this);
}

#if defined(ZAF_DEFAULT_DRAW)
ZafEventType ZafWindowObject::Draw(const ZafEventStruct &,
	ZafEventType ccode)
{
	return (ccode);
}

ZafEventType ZafWindowObject::DrawBackground(ZafRegionStruct &region, ZafEventType ccode)
{
//	if (ccode == S_REDISPLAY || ccode == S_REDISPLAY_REGION)
	{
		// Fill the background.
		display->SetPalette(LogicalPalette(ZAF_PM_BACKGROUND, PaletteState()));
		display->Rectangle(region, 0, true);
	}

	return (ccode);
}

ZafEventType ZafWindowObject::DrawBorder(ZafRegionStruct &region,
	ZafEventType ccode)
{
	if (parent && parentDrawBorder)
		return (parent->DrawBorder(region, ccode));

	ZafPaletteStruct lightShadow = LogicalPalette(ZAF_PM_LIGHT_SHADOW, ZAF_PM_ANY_STATE);
	ZafPaletteStruct darkShadow = LogicalPalette(ZAF_PM_DARK_SHADOW, ZAF_PM_ANY_STATE);
	ZafPaletteStruct whiteShadow = lightShadow;
	ZafPaletteStruct blackShadow = darkShadow;

	// Copy the background colors to the foreground since display->Line() uses
	// the foreground color. The "shadow" palettes have different colors in
	// the foreground and background entries so they can generate two shades
	// of shadow.
	whiteShadow.colorForeground = whiteShadow.colorBackground;
	blackShadow.colorForeground = blackShadow.colorBackground;

	display->SetPalette(darkShadow);
	display->Line(region.left, region.top,	region.right - 1, region.top, 1);
	display->Line(region.left, region.top + 1, region.left, region.bottom - 1, 1);

	display->SetPalette(blackShadow);
	display->Line(region.left + 1, region.top + 1, region.right - 1, region.top + 1, 1);
	display->Line(region.left + 1, region.top + 1, region.left + 1, region.bottom - 1, 1);

	display->SetPalette(whiteShadow);
	display->Line(region.right, region.top, region.right, region.bottom, 1);
	display->Line(region.left, region.bottom, region.right, region.bottom, 1);

	display->SetPalette(lightShadow);
	display->Line(region.right - 1, region.top + 1, region.right - 1, region.bottom - 1, 1);
	display->Line(region.left + 1, region.bottom - 1, region.right - 1, region.bottom - 1, 1);

	region -= ZAF_BORDER_SIZE;
	return (ccode);
}

ZafEventType ZafWindowObject::DrawFocus(ZafRegionStruct &,
	ZafEventType ccode)
{
	if (!SystemObject())
	{
		ZafRegionStruct focusRegion = ConvertToDrawRegion(this, &zafRegion);
		return (parent->DrawFocus(focusRegion, ccode));
	}

	return (ccode);
}

ZafEventType ZafWindowObject::DrawShadow(ZafRegionStruct &region,
	int depth, ZafEventType ccode)
{
	// Make sure there is depth to the object
	if (!depth)
		return (ccode);

	ZafPaletteStruct lightShadow = LogicalPalette(ZAF_PM_LIGHT_SHADOW, ZAF_PM_ANY_STATE);
	ZafPaletteStruct darkShadow = LogicalPalette(ZAF_PM_DARK_SHADOW, ZAF_PM_ANY_STATE);
	ZafPaletteStruct whiteShadow = lightShadow;
	ZafPaletteStruct blackShadow = darkShadow;

	// Copy the background colors to the foreground since display->Line() uses
	// the foreground color. The "shadow" palettes have different colors in
	// the foreground and background entries so they can generate two shades
	// of shadow.
	whiteShadow.colorForeground = whiteShadow.colorBackground;
	blackShadow.colorForeground = blackShadow.colorBackground;

	// Draw the shadow.
	int absoluteDepth = ZafAbs(depth);
	int halfDepth = absoluteDepth / 2;
	for (int i = 0; i < absoluteDepth; i++)
	{
		if (depth > 0)
			display->SetPalette(i > halfDepth ? lightShadow : whiteShadow);
		else
			display->SetPalette(i > halfDepth ? blackShadow : darkShadow);
		display->Line(region.left, region.top,	region.right - 1, region.top, 1);
		display->Line(region.left, region.top + 1, region.left, region.bottom - 1, 1);

		if (depth > 0)
			display->SetPalette(i > halfDepth ? blackShadow : darkShadow);
		else
			display->SetPalette(i > halfDepth ? lightShadow : whiteShadow);
		display->Line(region.right, region.top, region.right, region.bottom, 1);
		display->Line(region.left, region.bottom, region.right, region.bottom, 1);

		--region;
	}

	return (ccode);
}
#endif

ZafEventType ZafWindowObject::DragDropEvent(const ZafEventStruct &event)
{
	ZafEventType ccode = event.type;

	switch (ccode)
	{
	case S_DRAG_DEFAULT:
		ccode = S_DRAG_MOVE;
		// Fall through to S_DRAG_MOVE.
	case S_DRAG_MOVE:
	case S_DRAG_COPY:
		if (!AcceptDrop() && windowManager->dragObject != this)
		{
			eventManager->SetDeviceImage(E_MOUSE, DM_CANCEL);
			ccode = S_ERROR;
		}
		else
			eventManager->SetDeviceImage(E_MOUSE, ccode == S_DRAG_MOVE ?
				DM_DRAG_MOVE : DM_DRAG_COPY);
		break;

	case S_DROP_DEFAULT:
		ccode = S_DROP_MOVE;
		// Fall through to S_DROP_MOVE.
	case S_DROP_MOVE:
	case S_DROP_COPY:
		if (Disabled() || !AcceptDrop() || windowManager->dragObject == this)
			return (S_ERROR);
		else
		{
			const ZafIChar *text = ZAF_NULLP(ZafIChar);
			ZafWindowObject *dragObject = windowManager->dragObject;

			text = dragObject->Text();
			if (text)
				SetText(text);
			if (ccode == S_DROP_MOVE)
				dragObject->SetText(ZafLanguageManager::blankString);
		}
		break;

	case S_BEGIN_DRAG:
		windowManager->CaptureMouse(this);
		windowManager->dragObject = this;
		break;

	case S_END_DRAG:
		windowManager->CaptureMouse(ZAF_NULLP(ZafWindowObject));
		windowManager->dragObject = ZAF_NULLP(ZafWindowObject);
		eventManager->SetDeviceImage(E_MOUSE, DM_VIEW);
		break;

	default:
		ccode = S_ERROR;
		break;
	}

	return (ccode);
}

void ZafWindowObject::EndDraw(void)
{
	display->EndDraw();
}

ZafEventType ZafWindowObject::Event(const ZafEventStruct &event)
{
	static clock_t lastTime = clock(), currentTime = clock();
	static ZafWindowObject *lastSelected = ZAF_NULLP(ZafWindowObject);

	ZafEventType ccode = LogicalEvent(event);

	if (ccode != L_BEGIN_SELECT && modifyMode)
		Modify(event);
//		return (Modify(event));

	// Check for zinc events.
	switch (ccode)
	{
	// ----- Create messages -----
	case S_INITIALIZE:
		{
		// Convert "zafRegion" to OS (pixel) coordinates.
		modifyMode = 0;
#if defined(ZAF_GRAPHICS_MODE)
		ConvertCoordinates(ZAF_PIXEL);//(parent && RegionType() != ZAF_AVAILABLE_REGION) ? false : true);
#elif defined(ZAF_TEXT_MODE)
		ConvertCoordinates(ZAF_CELL);
#endif
		oldRegion.left = oldRegion.top = oldRegion.right = oldRegion.bottom = -10000;
		oldRegion.coordinateType = zafRegion.coordinateType;

		ZafWindowObject *rootObject = this;
		while (rootObject->parent)
			rootObject = rootObject->parent;

		// Initialize numberID and screenID.
		if (!numberID)
		{
			if (parent)
			{
				numberID = rootObject->NumberID();
				rootObject->SetNumberID((ZafNumberID)(numberID + 1));
			}
			else
			{
				numberID = windowManager->NumberID();
				windowManager->SetNumberID((ZafNumberID)(numberID + 1));
			}
		}
		// Make sure the root object always has the largest numberID.
		else if (parent && numberID >= rootObject->NumberID())
			rootObject->SetNumberID((ZafNumberID)(numberID + 1));
		// Set the stringID.
		if (!stringID)
		{
			ZafIChar newStringID[ZAF_MAXNAMELEN];
			strcpy(newStringID, parent ? genericFieldID : ClassName());
			itoa(numberID, &newStringID[strlen(newStringID)], 10, 0);
			SetStringID(newStringID);
		}
		}
		break;

	case S_DEINITIALIZE:
		if (windowManager->mouseObject == this)
			windowManager->mouseObject = ZAF_NULLP(ZafWindowObject);
		if (windowManager->MouseEventRoute() == this)
			windowManager->CaptureMouse(ZAF_NULLP(ZafWindowObject));
		if (windowManager->cursorObject == this)
			windowManager->cursorObject = ZAF_NULLP(ZafWindowObject);
		screenID = 0;
		break;

	case S_REGISTER_OBJECT:
		OSRegisterObject();
		break;

	case S_CREATE:
		{
		// Don't allow the object to be displayed until the creation
		// process is finished.
		bool saveVisible = visible;
		SetVisible(false);

		// Initialize member variables.
		Event(S_INITIALIZE);

		// Create the object with the OS.
		Event(S_REGISTER_OBJECT);

		// Allow the object to update its size/position (zafRegion).
		Event(S_COMPUTE_SIZE);

		// Size the object with the OS.
		OSSize();

		// Display the object on the screen (if visible).
		SetVisible(saveVisible);
		}
		break;

	// ----- Activation messages -----
	case S_CURRENT:
		focus = true;
		break;

	case S_NON_CURRENT:
		focus = false;
		break;

	case N_CURRENT:
	case N_NON_CURRENT:
		// Call the associated user function.
		ccode = (this->*memberUserFunction)(event, ccode);
		break;

	case S_SIZE_MODE:
	case S_MOVE_MODE:
		{
			ZafEventStruct sEvent;
			sEvent = event;
			if (ccode != sEvent.type)
			{
				sEvent.type = ccode;
				sEvent.rawCode = 0;
			}
			Modify(sEvent);
		}
		break;

	case S_REDISPLAY_REGION:
		if (AutomaticUpdate() && Visible() && screenID)
		{
			// When an object is sent a display message, the region passed is
			// relative to the object's top-left corner being (0,0).
			// So compare the drawing region to the object's (0,0)-based region.
			ZafRegionStruct region;
			region.left = region.top = 0;
			region.right = zafRegion.Width() - 1;
			region.bottom = zafRegion.Height() - 1;
			region.coordinateType = zafRegion.coordinateType;
			if (region.Overlap(event.region))
			{
				region = event.region;
				OSInvalidateRegion(&region, false, true);
//				Draw(event, ccode);
			}
		}
		break;

	case S_REDISPLAY_DATA:
	case S_REDISPLAY:
		if (AutomaticUpdate() && Visible() && screenID)
		{
			ZafRegionStruct region;
			region.left = region.top = 0;
			region.right = zafRegion.Width() - 1;
			region.bottom = zafRegion.Height() - 1;
			region.coordinateType = zafRegion.coordinateType;

			if (ccode == S_REDISPLAY)
				OSInvalidateRegion(&region, true, false);
			else
			{
				ZafEventStruct drawEvent(ccode);
				drawEvent.region = region;
				Draw(drawEvent, ccode);
			}
		}
		break;

	case S_SET_FOCUS:
		{
		ZafEventStruct focusEvent = event;
		focusEvent.windowObject = this;
		eventManager->Put(focusEvent);
		}
		break;

	// ----- Actions -----
	case L_VIEW:
		{
			if (!editMode)
				break;

			// Determine the type of modify operation.
			ZafRawCode sizeOperation = 0;
			if (event.position.column < ZAF_EDIT_MARGIN_SIZE)
				sizeOperation |= ZAF_SIZE_LEFT;
			else if (event.position.column >= zafRegion.Width() - ZAF_EDIT_MARGIN_SIZE)
				sizeOperation |= ZAF_SIZE_RIGHT;
			if (event.position.line < ZAF_EDIT_MARGIN_SIZE)
				sizeOperation |= ZAF_SIZE_TOP;
			else if (event.position.line >= zafRegion.Height() - ZAF_EDIT_MARGIN_SIZE)
				sizeOperation |= ZAF_SIZE_BOTTOM;

			// Set the mouse cursor image.
			if (sizeOperation & ZAF_SIZE_LEFT && sizeOperation & ZAF_SIZE_TOP)
				eventManager->SetDeviceImage(E_MOUSE, DM_TOP_LEFT_CORNER);
			else if (sizeOperation & ZAF_SIZE_RIGHT && sizeOperation & ZAF_SIZE_BOTTOM)
				eventManager->SetDeviceImage(E_MOUSE, DM_BOTTOM_RIGHT_CORNER);
			else if (sizeOperation & ZAF_SIZE_RIGHT && sizeOperation & ZAF_SIZE_TOP)
				eventManager->SetDeviceImage(E_MOUSE, DM_TOP_RIGHT_CORNER);
			else if (sizeOperation & ZAF_SIZE_LEFT && sizeOperation & ZAF_SIZE_BOTTOM)
				eventManager->SetDeviceImage(E_MOUSE, DM_BOTTOM_LEFT_CORNER);
			else if (sizeOperation & ZAF_SIZE_LEFT)
				eventManager->SetDeviceImage(E_MOUSE, DM_LEFT_SIDE);
			else if (sizeOperation & ZAF_SIZE_RIGHT)
				eventManager->SetDeviceImage(E_MOUSE, DM_RIGHT_SIDE);
			else if (sizeOperation & ZAF_SIZE_TOP)
				eventManager->SetDeviceImage(E_MOUSE, DM_TOP_SIDE);
			else if (sizeOperation & ZAF_SIZE_BOTTOM)
				eventManager->SetDeviceImage(E_MOUSE, DM_BOTTOM_SIDE);
			else
				eventManager->SetDeviceImage(E_MOUSE, DM_MOVE);
		}
		break;

	case L_BEGIN_SELECT:
		lastSelected = this;
		if (editMode)
		{
			// Modify (move / size) the object.
			Modify(event);

			ZafEventStruct editEvent = event;

			// Notify of active object.
			editEvent.type = D_SET_OBJECT;
			editEvent.rawCode = ClassID();
			editEvent.windowObject = this;
			Event(editEvent);

			// Notify of current position.
			editEvent.type = D_SET_POSITION;
			editEvent.rawCode = event.rawCode;
			editEvent.position = event.position;
			Event(editEvent);
		}
		else if (Draggable())
		{
			modifyRegion.left = event.position.column;
			modifyRegion.top = event.position.line;
		}
		break;

	case L_CONTINUE_SELECT:
//	case L_CONTINUE_COPY_DRAG:
//	case L_CONTINUE_LINK_DRAG:
		if (windowManager->dragObject)
		{
			if (acceptDrop || windowManager->dragObject == this)
			{
				ZafEventStruct dragEvent = event;
				dragEvent.type = S_DRAG_DEFAULT;

				// dragObject may not be draggable if it is in a combo box or tree.
				// Need to check its parent in these cases.
				ZafWindowObject *dragObject = windowManager->dragObject->Draggable() ?
					windowManager->dragObject : windowManager->dragObject->parent;
				if (ccode == L_CONTINUE_SELECT && dragObject->MoveDraggable())
					dragEvent.type = S_DRAG_MOVE;
/* START BLOCK COMMENT
**					else if (ccode == L_CONTINUE_COPY_DRAG && dragObject->CopyDraggable())
**						dragEvent.type = S_DRAG_COPY;
**					else if (ccode == L_CONTINUE_LINK_DRAG && dragObject->LinkDraggable())
**						dragEvent.type = S_DRAG_LINK;
END BLOCK COMMENT */
				else if (dragObject->MoveDraggable())
					dragEvent.type = S_DRAG_MOVE;
				else if (dragObject->CopyDraggable())
					dragEvent.type = S_DRAG_COPY;
				else if (dragObject->LinkDraggable())
					dragEvent.type = S_DRAG_LINK;

//				ccode = DragDropEvent(dragEvent);
				ccode = windowManager->DispatchMouseEvent(dragEvent);
				if (ccode == S_ERROR)
					eventManager->SetDeviceImage(E_MOUSE, DM_CANCEL);
			}
			else
			{
				eventManager->SetDeviceImage(E_MOUSE, DM_CANCEL);
				ccode = S_ERROR;
			}
		}
		else if (Draggable() && (ZafAbs(modifyRegion.left - event.position.column) >= 4 ||
			ZafAbs(modifyRegion.top - event.position.line) >= 4) && lastSelected == this)
			Event(S_BEGIN_DRAG);
		break;

	case L_END_SELECT:
//	case L_END_COPY_DRAG:
//	case L_END_LINK_DRAG:
		{
			ZafPositionStruct mousePosition = event.position;
			mousePosition.line += zafRegion.top;
			mousePosition.column += zafRegion.left;
			if (lastSelected != this && !windowManager->dragObject)
				break;

			currentTime = clock();

			if (windowManager->dragObject)
			{
//				if (acceptDrop && windowManager->dragObject != this)
				{
					ZafEventStruct dragEvent = event;
					dragEvent.type = S_DRAG_DEFAULT;

					bool copy = false, link = false, move = false;
					if ((event.modifiers | (S_CTRL | S_SHIFT)) == event.modifiers)
						link = true;
					else if (event.modifiers & S_CTRL)
						copy = true;
					else if (event.modifiers & S_SHIFT)
						move = true;

					// dragObject may not be draggable if it is in a combo box or
					// tree.  Need to check its parent in these cases.
					ZafWindowObject *dragObject = windowManager->dragObject->Draggable() ?
						windowManager->dragObject : windowManager->dragObject->parent;
					if (move && dragObject->MoveDraggable())
						dragEvent.type = S_DROP_MOVE;
/* START BLOCK COMMENT
**					else if (ccode == L_END_COPY_DRAG && dragObject->CopyDraggable())
**						dragEvent.type = S_DROP_COPY;
**					else if (ccode == L_END_LINK_DRAG && dragObject->LinkDraggable())
**						dragEvent.type = S_DROP_LINK;
END BLOCK COMMENT */
					else if (copy && dragObject->CopyDraggable())
						dragEvent.type = S_DROP_COPY;
					else if (link && dragObject->LinkDraggable())
						dragEvent.type = S_DROP_LINK;
					else
						dragEvent.type = S_DROP_DEFAULT;

//					ccode = DragDropEvent(dragEvent);
					ccode = windowManager->DispatchMouseEvent(dragEvent);
					Event(S_END_DRAG);
				}
				lastSelected = ZAF_NULLP(ZafWindowObject);
				break;
			}
			lastTime = currentTime;
			lastSelected = this;
		}
		break;

	case N_MOUSE_ENTER:
		// If help tip, update the help tips device.
		if ((QuickTip() || HelpObjectTip()))
			eventManager->Event(event, E_HELPTIPS);
		eventManager->SetDeviceState(E_MOUSE, DM_VIEW);
		break;

	case N_MOUSE_LEAVE:
		// If help tip, update the help tips device.
		if ((QuickTip() || HelpObjectTip()))
			eventManager->Event(event, E_HELPTIPS);
		break;

	// ----- Scroll messages -----
	case N_VSCROLL:
	case N_HSCROLL:
	case S_VSCROLL:
	case S_HSCROLL:
	case S_VSCROLL_SET:
	case S_HSCROLL_SET:
	case S_VSCROLL_CHECK:
	case S_HSCROLL_CHECK:
	case S_VSCROLL_COMPUTE:
	case S_HSCROLL_COMPUTE:
		ccode = ScrollEvent(event);
		break;

	// ----- Position & size messages -----
	case S_SIZE:
		SetRegion(event.region);
		ConvertCoordinates(ZAF_PIXEL);
/* START BLOCK COMMENT
**			SetConvertedCoordinate(false);
**			ConvertRegion((parent && RegionType() != ZAF_AVAILABLE_REGION) ? false : true);
END BLOCK COMMENT */
		Event(S_COMPUTE_SIZE);
		OSSize();
		break;

	case N_MOVE:
	case N_SIZE:
		ccode = 0;
		break;

	case S_COMPUTE_SIZE:
		if (RegionType() != ZAF_INSIDE_REGION)
  			SetRegion(parent ? parent->MaxRegion(this) : windowManager->MaxRegion(this));
		break;

	// ----- Drag & Drop -----
	case S_DRAG_MOVE:
	case S_DRAG_COPY:
	case S_DRAG_LINK:
	case S_DROP_MOVE:
	case S_DROP_COPY:
	case S_DROP_LINK:
	case S_DRAG_DEFAULT:
	case S_DROP_DEFAULT:
	case S_BEGIN_DRAG:
	case S_END_DRAG:
		ccode = DragDropEvent(event);
		break;

	case S_DESTROY:
		// Remove focus before destruction to allow data to be saved.
		if (focus)
		{
			windowManager->focusObject = ZAF_NULLP(ZafWindowObject);
			NotifyFocus(ZAF_NULLP(ZafWindowObject), false);
		}

		// Deinitialize the object.
		Event(ZafEventStruct(S_DEINITIALIZE));
		break;

	// ----- Miscellaneous -----
	case S_HELP:
	case L_HELP:
		if (helpContext && zafHelpSystem)
			zafHelpSystem->DisplayHelp(helpContext);
		else
		{
			// Allow the parent to display a help context if it has one.
			ccode = S_UNKNOWN;
		}
		break;

	case E_MOUSE:
		if (!Draggable())
			break;

		// Begin select
		if (event.rawCode == (M_LEFT | M_LEFT_CHANGE))
		{
			lastSelected = this;
			modifyRegion.left = event.position.column;
			modifyRegion.top = event.position.line;
		}
		// End and Continue select
		else if (windowManager->dragObject == this)
		{
			// Determine the requested operation.
			ZafEventStruct dragEvent = event;
			bool copy = false, link = false, move = false;
			if ((event.modifiers | (S_CTRL | S_SHIFT)) == event.modifiers)
				link = true;
			else if (event.modifiers & S_CTRL)
				copy = true;
			else if (event.modifiers & S_SHIFT)
				move = true;

			if (link && windowManager->dragObject->LinkDraggable())
				dragEvent.type = (event.rawCode == M_LEFT) ? S_DRAG_LINK : S_DROP_LINK;
			else if (copy && windowManager->dragObject->CopyDraggable())
				dragEvent.type = (event.rawCode == M_LEFT) ? S_DRAG_COPY : S_DROP_COPY;
			else if (move && windowManager->dragObject->MoveDraggable())
				dragEvent.type = (event.rawCode == M_LEFT) ? S_DRAG_MOVE : S_DROP_MOVE;
			else
				dragEvent.type = (event.rawCode == M_LEFT) ? S_DRAG_DEFAULT : S_DROP_DEFAULT;

			if (windowManager->DispatchMouseEvent(dragEvent) == S_ERROR)
				eventManager->SetDeviceImage(E_MOUSE, DM_CANCEL);

			if (event.rawCode == M_LEFT_CHANGE)
				Event(S_END_DRAG);
		}
		else if (windowManager->dragObject) // Shouldn't get here!
			DragDropEvent(event);
		else if (lastSelected == this && (ZafAbs(modifyRegion.left - event.position.column) >= 4 ||
			ZafAbs(modifyRegion.top - event.position.line) >= 4) && event.rawCode == M_LEFT)
			Event(S_BEGIN_DRAG);
		break;

	// ----- Unknown or error messages -----
	default:
		ccode = S_UNKNOWN;
	}

	// Return the control code.
	return (ccode);
}

void ZafWindowObject::GetClip(ZafWindowObject *,
	ZafRegionStruct &clip, ZafRegionStruct &virtualRegion)
{
	// Let the parent adjust the regions first.
	if (parent)
		parent->GetClip(this, clip, virtualRegion);

	// Adjust the virtualRegion region.
	virtualRegion.left += zafRegion.left;
	virtualRegion.top += zafRegion.top;
	virtualRegion.right += zafRegion.left;
	virtualRegion.bottom += zafRegion.top;

	// Adjust the clip region.
	if (virtualRegion.left > clip.left)
		clip.left = virtualRegion.left;
	if (virtualRegion.top > clip.top)
		clip.top = virtualRegion.top;
	clip.right = ZafMin(clip.right, clip.left + zafRegion.Width() - 1);
	clip.bottom = ZafMin(clip.bottom, clip.top + zafRegion.Height() - 1);
}

ZafRegionStruct ZafWindowObject::MaxRegion(ZafWindowObject *, ZafVtJustify, ZafHzJustify)
{
	// Default is the object's region.
	ZafRegionStruct region;
	region.left = region.top = 0;
	region.right = zafRegion.right - zafRegion.left;
	region.bottom = zafRegion.bottom - zafRegion.top;
	region.coordinateType = zafRegion.coordinateType;

	return (region);
}

ZafEventType ZafWindowObject::Modify(const ZafEventStruct &event)
{
	ZafEventType ccode = LogicalEvent(event);
	static ZafRawCode sizeOperation = 0;

	static ZafPaletteStruct rubberBandPalette =
	{
		ZAF_LINE_SOLID, ZAF_PTN_SOLID_FILL, ZAF_CLR_LIGHTCYAN, ZAF_CLR_BLACK, ZAF_MONO_NULL, ZAF_MONO_NULL, ZAF_FNT_DIALOG
	};
	ZafPaletteStruct *oldPalette = ((ZafScreenDisplay *)display)->xorPalette;

	// Save the original mouse coordinates.
	static ZafPositionStruct originalPosition;
	ZafRegionStruct osModifyRegion;
	ZafRegionStruct osLastRegion;
//	ZafEventType mouseState = eventManager->Event(D_STATE, E_MOUSE);

	switch (ccode)
	{
		case S_SIZE_MODE:
			{
				windowManager->CaptureMouse(this);

				// Make sure all damage has been repaired so that the
				// rubber-band outline appears correctly.
				windowManager->RepairDamage();

				// Draw the initial xor rectangle.
				modifyRegion = zafRegion;
				lastRegion = modifyRegion;
				if (parent)
					osModifyRegion = parent->ConvertToOSRegion(this, &modifyRegion);
				else
					osModifyRegion = modifyRegion;

//				originalPosition = event.position;
				originalPosition.line = event.osEvent.position.line;
				originalPosition.column = event.osEvent.position.column;
				originalPosition.coordinateType = event.position.coordinateType;

//				eventManager->SetDeviceState(E_MOUSE, D_HIDE);
				ZafRegionStruct clip;
				ZafRegionStruct draw;
#if defined(ZAF_GRAPHICS_MODE)
				clip.Assign(0, 0, display->columns, display->lines);
				clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
				clip.Assign(0, 0, display->columns - 1, display->lines - 1);
				clip.coordinateType = ZAF_CELL;
#endif
				draw = clip;
				if (parent)
					parent->GetClip(this, clip, draw);
				display->BeginDraw(0, ID_ZAF_SCREEN, draw, clip);
				display->SetClipRegion(clip);
				((ZafScreenDisplay *)display)->xorPalette = &rubberBandPalette;
				display->RectangleXORDiff(ZAF_NULLP(ZafRegionStruct), &osModifyRegion);
				display->EndDraw();
				((ZafScreenDisplay *)display)->xorPalette = oldPalette;
//				eventManager->Event(mouseState, E_MOUSE);
				modifyMode = S_SIZE_MODE;
				if (!sizeOperation)
					sizeOperation = event.rawCode;
				if (!sizeOperation)
					sizeOperation = ZAF_SIZE_RIGHT | ZAF_SIZE_BOTTOM;
			}
			break;

		case S_MOVE_MODE:
			{
				windowManager->CaptureMouse(this);

				// Make sure all damage has been repaired so that the
				// rubber-band outline appears correctly.
				windowManager->RepairDamage();

				lastRegion = modifyRegion = zafRegion;
#if 1
				if (parent)
					osModifyRegion = parent->ConvertToOSRegion(this, &modifyRegion);
				else
					osModifyRegion = modifyRegion;

//				originalPosition = event.position;
				originalPosition.line = event.osEvent.position.line;
				originalPosition.column = event.osEvent.position.column;
				originalPosition.coordinateType = event.position.coordinateType;

//				eventManager->SetDeviceState(E_MOUSE, D_HIDE);
				eventManager->SetDeviceState(E_CURSOR, D_OFF);
				ZafRegionStruct clip;
				ZafRegionStruct draw;
#if defined(ZAF_GRAPHICS_MODE)
				clip.Assign(0, 0, display->columns, display->lines);
				clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
				clip.Assign(0, 0, display->columns - 1, display->lines - 1);
				clip.coordinateType = ZAF_CELL;
#endif
				draw = clip;
				if (parent)
					parent->GetClip(this, clip, draw);
				display->BeginDraw(0, ID_ZAF_SCREEN, draw, clip);
				display->SetClipRegion(clip);
				((ZafScreenDisplay *)display)->xorPalette = &rubberBandPalette;
				display->RectangleXORDiff(ZAF_NULLP(ZafRegionStruct), &osModifyRegion);
				display->EndDraw();
				((ZafScreenDisplay *)display)->xorPalette = oldPalette;
//				eventManager->Event(mouseState, E_MOUSE);
#else
//				originalPosition = event.position;
				originalPosition.line = event.osEvent.position.line;
				originalPosition.column = event.osEvent.position.column;
				originalPosition.coordinateType = event.position.coordinateType;
#endif
				sizeOperation = ZAF_SIZE_LEFT | ZAF_SIZE_RIGHT | ZAF_SIZE_TOP | ZAF_SIZE_BOTTOM;
				modifyMode = S_MOVE_MODE;
			}
			break;

		case L_BEGIN_SELECT:
			windowManager->CaptureMouse(this);

			// Position coordinates are converted to 0,0 based.
//			originalPosition = event.position;
			originalPosition.line = event.osEvent.position.line;
			originalPosition.column = event.osEvent.position.column;
			originalPosition.coordinateType = event.position.coordinateType;

			lastRegion = modifyRegion = zafRegion;
			sizeOperation = 0;

			// Determine the type of modify operation.
			if (event.position.column <= ZAF_EDIT_MARGIN_SIZE)
				sizeOperation |= ZAF_SIZE_LEFT;
			else if (event.position.column >= zafRegion.Width() - ZAF_EDIT_MARGIN_SIZE)
				sizeOperation |= ZAF_SIZE_RIGHT;
			if (event.position.line <= ZAF_EDIT_MARGIN_SIZE)
				sizeOperation |= ZAF_SIZE_TOP;
			else if (event.position.line >= zafRegion.Height() - ZAF_EDIT_MARGIN_SIZE)
				sizeOperation |= ZAF_SIZE_BOTTOM;

			// Setting all flags acts as a move operation.
			if (!sizeOperation)
			{
				sizeOperation = ZAF_SIZE_LEFT | ZAF_SIZE_RIGHT | ZAF_SIZE_TOP | ZAF_SIZE_BOTTOM;
				modifyMode = S_MOVE_MODE;
				eventManager->SetDeviceState(E_CURSOR, D_HIDE);
			}
			else
			{
				modifyMode = S_SIZE_MODE;
				if (parent)
					osModifyRegion = parent->ConvertToOSRegion(this, &modifyRegion);
				else
					osModifyRegion = modifyRegion;

				// Draw the initial xor rectangle.
//				eventManager->SetDeviceState(E_MOUSE, D_HIDE);
				ZafRegionStruct clip;
				ZafRegionStruct draw;
#if defined(ZAF_GRAPHICS_MODE)
				clip.Assign(0, 0, display->columns, display->lines);
				clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
				clip.Assign(0, 0, display->columns - 1, display->lines - 1);
				clip.coordinateType = ZAF_CELL;
#endif
				draw = clip;
				if (parent)
					parent->GetClip(this, clip, draw);
				display->BeginDraw(0, ID_ZAF_SCREEN, draw, clip);
				display->SetClipRegion(clip);
				((ZafScreenDisplay *)display)->xorPalette = &rubberBandPalette;
				display->RectangleXORDiff(ZAF_NULLP(ZafRegionStruct), &osModifyRegion);
				display->EndDraw();
				((ZafScreenDisplay *)display)->xorPalette = oldPalette;
//				eventManager->Event(mouseState, E_MOUSE);
			}
			break;

		case L_CONTINUE_SELECT:
			{
//				ZafPositionStruct deltaPosition = event.position;
				ZafPositionStruct deltaPosition;
				deltaPosition.line  = event.osEvent.position.line;
				deltaPosition.column  = event.osEvent.position.column;
				deltaPosition.coordinateType = event.position.coordinateType;
				// Constrain the mouse.
				ZafRegionStruct clipRegion;
#if defined(ZAF_GRAPHICS_MODE)
				clipRegion.Assign(0, 0, display->columns, display->lines);
				clipRegion.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
				clipRegion.Assign(0, 0, display->columns - 1, display->lines - 1);
				clipRegion.coordinateType = ZAF_CELL;
#endif
				ZafRegionStruct virtualRegion = clipRegion;
				if (parent)
					parent->GetClip(this, clipRegion, virtualRegion);
				if (parent)
				{
//					ZafPositionStruct osDeltaPosition = ConvertToOSPosition(this, &deltaPosition);
					ZafPositionStruct osDeltaPosition = deltaPosition;
					if (osDeltaPosition.line < clipRegion.top)
					{
						deltaPosition.line += clipRegion.top - osDeltaPosition.line;
						osDeltaPosition.line = clipRegion.top;
					}
					if (osDeltaPosition.line > clipRegion.bottom)
					{
						deltaPosition.line += clipRegion.bottom - osDeltaPosition.line;
						osDeltaPosition.line = clipRegion.bottom;
					}
					if (osDeltaPosition.column < clipRegion.left)
					{
						deltaPosition.column += clipRegion.left - osDeltaPosition.column;
						osDeltaPosition.column = clipRegion.left;
					}
					if (osDeltaPosition.column > clipRegion.right)
					{
						deltaPosition.column += clipRegion.right - osDeltaPosition.column;
						osDeltaPosition.column = clipRegion.right;
					}
					if (!SupportObject())
						eventManager->SetDevicePosition(E_MOUSE, osDeltaPosition.column,
							osDeltaPosition.line);
/* START BLOCK COMMENT
**						int deltaY = deltaPosition.line - originalPosition.line;
**						int deltaX = deltaPosition.column - originalPosition.column;
**						if (modifyRegion.left + deltaX < 0)
**							deltaPosition.column -= modifyRegion.left + deltaX;
**						if (modifyRegion.top + deltaY < 0)
**							deltaPosition.line -= modifyRegion.top + deltaY;
END BLOCK COMMENT */
				}

				deltaPosition.column -= originalPosition.column;
				deltaPosition.line -= originalPosition.line;

				// Snap the position to the proper coordinates.
				if (editMode)
				{
					deltaPosition.ConvertCoordinates(coordinateType);
					deltaPosition.ConvertCoordinates(zafRegion.coordinateType);
				}

				// Compute the new moving/sizeing region.
				if (SizeLeft(sizeOperation))
					modifyRegion.left += deltaPosition.column;
				if (SizeTop(sizeOperation))
					modifyRegion.top += deltaPosition.line;
				if (SizeRight(sizeOperation))
					modifyRegion.right += deltaPosition.column;
				if (SizeBottom(sizeOperation))
					modifyRegion.bottom += deltaPosition.line;

				// Don't let the object be sized too small.
				int minWindowWidth = 20 * display->cellWidth;
				int minWindowHeight = 4 * display->cellHeight;
				int minObjectWidth = display->cellWidth;
				int minObjectHeight = display->cellHeight;
				if (modifyMode == S_SIZE_MODE)
				{
					if ((!parent || IsA(ID_ZAF_MDI_WINDOW)) && modifyRegion.Width() < minWindowWidth)
					{
						if (SizeLeft(sizeOperation))
						{
							modifyRegion.left = modifyRegion.right - minWindowWidth + 1;
							deltaPosition.column = modifyRegion.left - lastRegion.left;
						}
						else if (SizeRight(sizeOperation))
						{
							modifyRegion.right = modifyRegion.left + minWindowWidth - 1;
							deltaPosition.column = modifyRegion.right - lastRegion.right;
						}
					}
					else if (parent && modifyRegion.Width() < minObjectWidth)
					{
						if (SizeLeft(sizeOperation))
						{
							modifyRegion.left = modifyRegion.right - minObjectWidth + 1;
							deltaPosition.column = modifyRegion.left - lastRegion.left;
						}
						else if (SizeRight(sizeOperation))
						{
							modifyRegion.right = modifyRegion.left + minObjectWidth - 1;
							deltaPosition.column = modifyRegion.right - lastRegion.right;
						}
					}

					if ((!parent || IsA(ID_ZAF_MDI_WINDOW)) && modifyRegion.Height() < minWindowHeight)
					{
						if (SizeTop(sizeOperation))
						{
							modifyRegion.top = modifyRegion.bottom - minWindowHeight + 1;
							deltaPosition.line = modifyRegion.top - lastRegion.top;
						}
						else if (SizeBottom(sizeOperation))
						{
							modifyRegion.bottom = modifyRegion.top + minWindowHeight - 1;
							deltaPosition.line = modifyRegion.bottom - lastRegion.bottom;
						}
					}
					else if (parent && modifyRegion.Height() < minObjectHeight)
					{
						if (SizeTop(sizeOperation))
						{
							modifyRegion.top = modifyRegion.bottom - minObjectHeight + 1;
							deltaPosition.line = modifyRegion.top - lastRegion.top;
						}
						else if (SizeBottom(sizeOperation))
						{
							modifyRegion.bottom = modifyRegion.top + minObjectHeight - 1;
							deltaPosition.line = modifyRegion.bottom - lastRegion.bottom;
						}
					}
				}

				// Move the xor rectangle.
				if (modifyRegion != lastRegion)
				{
					originalPosition.column += deltaPosition.column;
					originalPosition.line += deltaPosition.line;
#if 0
					// Calculate OS Positions first.
					if (modifyMode == S_SIZE_MODE)
#endif
					{
						if (parent)
						{
							osModifyRegion = parent->ConvertToOSRegion(this, &modifyRegion);
							osLastRegion = parent->ConvertToOSRegion(this, &lastRegion);
						}
						else
						{
							osModifyRegion = modifyRegion;
							osLastRegion = lastRegion;
						}
						((ZafScreenDisplay *)display)->xorPalette = &rubberBandPalette;

//						eventManager->SetDeviceState(E_MOUSE, D_HIDE);
						ZafRegionStruct clip;
						ZafRegionStruct draw;
#if defined(ZAF_GRAPHICS_MODE)
						clip.Assign(0, 0, display->columns, display->lines);
						clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
						clip.Assign(0, 0, display->columns - 1, display->lines - 1);
						clip.coordinateType = ZAF_CELL;
#endif
						draw = clip;
						if (parent)
							parent->GetClip(this, clip, draw);
						display->BeginDraw(0, ID_ZAF_SCREEN, draw, clip);
						display->SetClipRegion(clip);
						display->RectangleXORDiff(&osLastRegion, &osModifyRegion);
						display->EndDraw();
						((ZafScreenDisplay *)display)->xorPalette = oldPalette;
//						eventManager->Event(mouseState, E_MOUSE);
					}
#if 0
					else
					{
						// RegionCopy
						if (parent)
						{
							osLastRegion = lastRegion;
							parent->GetClip(this, clipRegion, osLastRegion);
							osModifyRegion = parent->ConvertToOSRegion(this, &modifyRegion);
						}
						else
						{
							osModifyRegion = modifyRegion;
							osLastRegion = lastRegion;
						}
						ZafRegionStruct copyRegion;
						clipRegion.Overlap(osLastRegion, copyRegion);
						if (osModifyRegion.bottom > clipRegion.bottom && deltaPosition.line > 0)
							copyRegion.bottom -= deltaPosition.line;
						if (osModifyRegion.right > clipRegion.right && deltaPosition.column > 0)
							copyRegion.right -= deltaPosition.column;

						if (clipRegion.Overlap(osModifyRegion))
							display->RegionCopy(copyRegion, osModifyRegion.left, osModifyRegion.top);
						ZafRegionList damageList;
						damageList.Add(new ZafRegionElement(0, osModifyRegion));
						damageList.Add(new ZafRegionElement(0, osLastRegion));
						clipRegion.Overlap(osModifyRegion, copyRegion);
						if (osLastRegion.bottom > clipRegion.bottom)
						{
							if (deltaPosition.line > 0)
								copyRegion.bottom -= deltaPosition.line;
							else
								copyRegion.bottom += deltaPosition.line;
						}
						if (osLastRegion.right > clipRegion.right)
						{
							if (deltaPosition.column > 0)
								copyRegion.right -= deltaPosition.column;
							else
								copyRegion.right += deltaPosition.column;
						}

						damageList.Split(0, copyRegion);
						zafRegion = modifyRegion;
						for (ZafRegionElement *regionElement = damageList.First();
							regionElement; regionElement = regionElement->Next())
								windowManager->OSInvalidateRegion(&regionElement->region);
						windowManager->RepairDamage();
					}
#endif
					lastRegion = modifyRegion;
				}
			}
			break;

		case L_END_SELECT:
			sizeOperation = 0;
			modifyMode = 0;
			windowManager->CaptureMouse(ZAF_NULLP(ZafWindowObject));
#if 0
			if (modifyMode == S_SIZE_MODE)
#endif
			{
				// Erase the xor rectangle.
				if (parent)
					osLastRegion = parent->ConvertToOSRegion(this, &lastRegion);
				else
					osLastRegion = lastRegion;
				((ZafScreenDisplay *)display)->xorPalette = &rubberBandPalette;
//				eventManager->SetDeviceState(E_MOUSE, D_HIDE);
				ZafRegionStruct clip;
				ZafRegionStruct draw;
#if defined(ZAF_GRAPHICS_MODE)
				clip.Assign(0, 0, display->columns, display->lines);
				clip.coordinateType = ZAF_PIXEL;
#elif defined(ZAF_TEXT_MODE)
				clip.Assign(0, 0, display->columns - 1, display->lines - 1);
				clip.coordinateType = ZAF_CELL;
#endif
				draw = clip;
				if (parent)
					parent->GetClip(this, clip, draw);
				display->BeginDraw(0, ID_ZAF_SCREEN, draw, clip);
				display->SetClipRegion(clip);
				display->RectangleXORDiff(&osLastRegion, ZAF_NULLP(ZafRegionStruct));
				display->EndDraw();
				((ZafScreenDisplay *)display)->xorPalette = oldPalette;
//				ZafEventStruct enterEvent = event;
//				enterEvent.type = N_MOUSE_ENTER;
//				enterEvent.type = L_VIEW;

				// Finalize the position.
				Event(ZafEventStruct(S_SIZE, 0, modifyRegion));
//				eventManager->Event(mouseState, E_MOUSE);
//				windowManager->DispatchMouseEvent(enterEvent);

/* START BLOCK COMMENT
**						if (windowManager->cursorObject &&
**							(windowManager->cursorObject->RootObject() == this ||
**							windowManager->cursorObject == this))
**							windowManager->cursorObject->Event(S_REDISPLAY_CURSOR);
END BLOCK COMMENT */
			}
			break;
	}

	// Return success.
	return (ccode);
}

ZafEventType ZafWindowObject::MoveEvent(const ZafEventStruct &)
{
	return (S_CONTINUE);
}

void ZafWindowObject::RegisterObject(void)
{
	// This code must be here in case a class (such as ZafTableRecord) defines
	// an OSRegisterObject() method that does not call
	// ZafWindowObject::OSRegisterObject() in order to not register the object
	// with the OS.
	if (!screenID)
	{
		// A non-SystemObject() uses the parent screenID.
		if (SystemObject())
			OSRegisterObject();
		else if (parent)
			screenID = parent->screenID;
	}
}

ZafEventType ZafWindowObject::ScrollEvent(const ZafEventStruct &)
{
	return (S_ERROR);
}

// ----- Attributes ---------------------------------------------------------

bool ZafWindowObject::SetAcceptDrop(bool acceptDrop)
{
	ZafWindowObject::acceptDrop = acceptDrop;
	return (acceptDrop);
}

bool ZafWindowObject::SetAutomaticUpdate(bool tAutomaticUpdate)
{
	if (automaticUpdate != tAutomaticUpdate)
	{
		automaticUpdate = tAutomaticUpdate;

		if (automaticUpdate && Visible())
			Redisplay();
	}

	return (automaticUpdate);
}

bool ZafWindowObject::SetDisabled(bool setDisabled)
{
	if (disabled != setDisabled)
	{
		disabled = setDisabled;
		Redisplay();
	}

	return (disabled);
}

bool ZafWindowObject::SetEditMode(bool setEditMode)
{
	// Make sure the attribute has changed.
	if (editMode != setEditMode && !screenID)
		editMode = setEditMode;

	// Return the current attribute.
	return (editMode);
}

bool ZafWindowObject::SetFocus(bool tFocus)
{
	// If the object can actually get focus, i.e., its parent has focus,
	// it will receive an S_CURRENT message from the parent's (or
	// windowManager's) NotifyFocus().  S_CURRENT will set the focus
	// boolean.  If the object cannot receive focus, it will be made current
	// in its parent's list so that it will get focus when the parent does.
/* START BLOCK COMMENT
**		if (Disabled() || Noncurrent())
**			focus = false;
**		if (focus == tFocus)
**			return (focus);
**	
**		if (tFocus)
**			NotifyFocus(this, true);
**		else
**			// Tell the object to display itself without focus.
**			Event(N_NON_CURRENT);
**	
**		return (focus);
END BLOCK COMMENT */

	if (Disabled() || Noncurrent())
		tFocus = false;

	if (focus != tFocus)
	{
		// Propagate the focus.
		if (screenID)
		{
			if (tFocus)
				NotifyFocus(this, true);
			else
			{
				// Tell the object to display itself without focus.
				NotifyFocus(parent, false);
//				if (parent)
//					DynamicPtrCast(parent, ZafWindow)->SetCurrent(ZAF_NULLP(ZafWindowObject));
			}
		}
//		focus = tFocus;
	}

	return (focus);
}

bool ZafWindowObject::SetNoncurrent(bool setNoncurrent)
{
	// Make sure the attribute has changed.
	if (noncurrent != setNoncurrent)
		noncurrent = setNoncurrent;

	// Return the current attribute.
	return (noncurrent);
}

bool ZafWindowObject::SetVisible(bool tVisible)
{
	if (visible != tVisible)
	{
		visible = tVisible;

		if (visible) // Draw.
		{
			if (parent && regionType == ZAF_AVAILABLE_REGION)
			{
				parent->OSSize();
				Redisplay();
			}
			else if (!parent)
				windowManager->OSInvalidateRegion(&zafRegion, true);
			else
				Redisplay();
		}
		else // Erase.
		{
			if (parent)
			{
				if (!SupportObject())
					parent->OSInvalidateRegion(&zafRegion, false, true);
				else if (regionType == ZAF_AVAILABLE_REGION)
					parent->OSSize();
				else
					parent->OSInvalidateRegion(&zafRegion, false, false);
			}
			else if (screenID)
			{
				// Have the window manager draw behind window.
				windowManager->OSInvalidateRegion(&zafRegion, true);
			}
		}
	}
	return (visible);
}

// ----- OS Specific Functions ----------------------------------------------

ZafEventType ZafWindowObject::OSEvent(const ZafEventStruct &)
{
	return (S_ERROR);
}

void ZafWindowObject::OSInvalidateRegion(ZafRegionStruct *region, bool immediateRedraw, bool)
{
	if (AutomaticUpdate() && Visible() && screenID)
	{
/* START BLOCK COMMENT
**			if (!immediateRedraw)
**			{
**				ZafRegionStruct tRegion = region ? *region : zafRegion;
**				if (!region)
**				{
**					tRegion.HzShift(-zafRegion.left);
**					tRegion.VtShift(-zafRegion.top);
**				}
**				ZafRegionStruct osRegion;
**				osRegion = ConvertToOSRegion(this, &tRegion);
**				windowManager->OSInvalidateRegion(&osRegion);
**				return;
**			}
END BLOCK COMMENT */

		// The parent is responsible for non system objects.
		if (SystemObject() || SupportObject())
		{
			ZafEventStruct event(S_REDISPLAY_REGION);
			if (region)
				event.region = *region;
			else
			{
				event.region = zafRegion;
				event.region.HzShift(-event.region.left);
				event.region.VtShift(-event.region.top);
			}
			Draw(event, S_REDISPLAY_REGION);
		}
		else if (parent)
		{
			ZafRegionStruct objectRegion;
			if (region)
			{
				objectRegion = *region;
				objectRegion.HzShift(zafRegion.left);
				objectRegion.VtShift(zafRegion.top);
			}
			else
				objectRegion = zafRegion;

			parent->OSInvalidateRegion(&objectRegion, immediateRedraw, true);
		}
	}
}

void ZafWindowObject::OSMapPalette(ZafPaletteStruct &palette, ZafPaletteType, ZafPaletteState)
{
	if (palette.lineStyle == ZAF_LINE_DEFAULT)
		palette.lineStyle = ZAF_LINE_NULL;
	if (palette.fillPattern == ZAF_PTN_DEFAULT)
		palette.fillPattern = ZAF_PTN_NULL;
	if (palette.colorForeground == ZAF_CLR_DEFAULT)
		palette.colorForeground = ZAF_CLR_NULL;
	if (palette.colorBackground == ZAF_CLR_DEFAULT)
		palette.colorBackground = ZAF_CLR_NULL;
	if (palette.monoForeground == ZAF_MONO_DEFAULT)
		palette.monoForeground = ZAF_MONO_NULL;
	if (palette.monoBackground == ZAF_MONO_DEFAULT)
		palette.monoBackground = ZAF_MONO_NULL;
	if (palette.font == ZAF_FNT_DEFAULT)
		palette.font = ZAF_FNT_NULL;
}

void ZafWindowObject::OSRegisterObject(void)
{
	if (!screenID)
	{
		if (!parent /*|| IsMDI() || Temporary()*/)
			screenID = ++(windowManager->currentScreenID);
		else
			screenID = parent->screenID;
	}
}

OSWindowID ZafWindowObject::OSScreenID(ZafScreenIDType) const
{
	return (screenID);
}

ZafError ZafWindowObject::OSSetText(void)
{
	// Setting ostext at window object level is invalid.
	return (ZAF_ERROR_INVALID);
}

void ZafWindowObject::OSSize(void)
{
	// Don't do anything if didn't change size.
	if (zafRegion == oldRegion)
		return;

	if (!screenID)
		return;

	// Redraw old region.
	if (parent && !SupportObject())
	{
		// Only invalidate the parts that don't overlap the new region.
		ZafRegionStruct clip;
		ZafRegionList damageList;
		if (zafRegion.Overlap(oldRegion))
		{
			damageList.Add(new ZafRegionElement(0, oldRegion));
			clip = zafRegion;
			damageList.Split(0, clip);
		}
		else
			damageList.Add(new ZafRegionElement(0, oldRegion));

		if (damageList.First())
		{
			for (ZafRegionElement *regionElement = damageList.First();
				regionElement; regionElement = regionElement->Next())
					parent->OSInvalidateRegion(&regionElement->region, false, true);
		}
/* START BLOCK COMMENT
**			ZafRegionStruct osUpdateRegion = parent->ConvertToOSRegion(this, &oldRegion);
**			windowManager->OSInvalidateRegion(&osUpdateRegion, false);
END BLOCK COMMENT */
	}

	oldRegion = zafRegion;

	if (AutomaticUpdate() && Visible())
	{
		OSInvalidateRegion(0, false, false);
/* START BLOCK COMMENT
**			ZafRegionStruct osUpdateRegion = parent ? parent->ConvertToOSRegion(this, &zafRegion) : zafRegion;
**			windowManager->OSInvalidateRegion(&osUpdateRegion, false);
END BLOCK COMMENT */
	}

	if (oldRegion.left != zafRegion.left || oldRegion.top != zafRegion.top)
		Event(N_MOVE);
	Event(N_SIZE);
}

ZafError ZafWindowObject::OSUpdatePalettes(ZafPaletteType , ZafPaletteType )
{
	if (screenID)
		Redisplay();
	return (ZAF_ERROR_NONE);
}

